import importlib
import logging
import numpy as np

from xframe.library import physicsLibrary as pLib
from xframe.library.pythonLibrary import convert_to_slice_if_possible
from xframe.library.pythonLibrary import split_into_simple_slices
from xframe.library.pythonLibrary import xprint

from xframe.interfaces import ExperimentWorkerInterface
from xframe.interfaces import CommunicationInterface

from .detectors.agipd import AGIPD
from .expLibrary import filters
from .expLibrary.filters import Filter,FilterSequence,FilterTools
from .expLibrary.rois import ROIManager
from .expLibrary.misk import Selection

from multiprocessing import cpu_count
from xframe import settings
from xframe import database
from xframe import Multiprocessing
from xframe.library.pythonLibrary import DictNamespace

log=logging.getLogger('root')

class DataSelection(DictNamespace):
    def __init__(self,run : int,frame_range=slice(None),cells=slice(None),cells_mode='relative',pulses=slice(None),pulses_mode='relative',trains=slice(None),trains_mode='relative',modules = np.arange(16),n_frames=20000,good_cells=np.arange(1,202),in_multiples_of=False):
        super().__init__()
        self['run'] = run
        self['frame_range']=frame_range
        
        selection = DictNamespace(cells=DictNamespace(range=cells,mode=cells_mode),
                                  pulses = DictNamespace(range=pulses,mode=pulses_mode),
                                  trains = DictNamespace(range=trains,mode=trains_mode)
                                  )
        self['selection']=selection
        self['modules']=modules
        self['n_frames']=n_frames
        self['good_cells']=good_cells
        self['in_multiples_of']=in_multiples_of

class ExperimentWorker(ExperimentWorkerInterface):
    DataSelection = DataSelection
    def __init__(self,detector=False,calibrator=False):
        opt = settings.experiment
        comm_module = Multiprocessing.comm_module
        self.opt=opt
        self.data_mode = opt.get('data_mode','proc')
        self.sample_distance = opt.get('sample_distance',700)# in mm
        self.x_ray_energy = opt.get('x_ray_energy',6010)# in eV
        self.x_ray_wavelength = pLib.energy_to_wavelength(self.x_ray_energy)
        self.comm_module=comm_module
        self.db = database.experiment
        if isinstance(detector,bool):
            detector = AGIPD(self.db)
        self.detector = detector
        #if isinstance(calibrator,bool):
        #    calibrator = AGIPD_VDS_Calibrator()
        #self.calibrator = calibrator
        self.info=self._generate_info()
        self.detector.origin = np.array([0,0,self.sample_distance])
        self.custom_mask = np.full(self.detector.data_shape,True)
        use_custom_mask = opt.get('custom_mask',False)
        if use_custom_mask:
            self.custom_mask = self.db.load('custom_mask')

        self.roi_manager = self.load_roi_manager()
        #self.rois = self.create_ROIs()

        #self.filter_sequence = self.load_filters()
        #background_opt = opt['calibrator']['subtract_background']
        #self.apply_background_subtraction = background_opt.get('apply',False)
        #self.background = 0
        #if self.apply_background_subtraction:
        #    self.background = self.db.load('background',path_modifiers={'name':background_opt['name']})['background']
        
        self.ids_by_run = {}
        self.cell_ids = None
        self.pulse_ids = None
        self.train_ids = None
        self.nframes = None


    ## get general information from experiment and run to be analyzed##
    def _generate_info(self):
        #data_info =self.db.load('info',runs=self.opt['runs'],data_mode=self.data_mode)
        #info = {**data_info,'sample_distance':self.sample_distance,'x_ray_energy':self.x_ray_energy,'x_ray_wavelength':self.x_ray_wavelength}
        #data_info =self.db.load('info',runs=self.opt['runs'],data_mode=self.data_mode)
        info = {'sample_distance':self.sample_distance,'x_ray_energy':self.x_ray_energy,'x_ray_wavelength':self.x_ray_wavelength}
        return info
    def set_ids_for_run(self,run):
        ids_by_run = self.ids_by_run
        run_str=str(run)
        if run_str in ids_by_run:
            self.cell_ids = ids_by_run[run_str]['cell_ids']
            self.pulse_ids = ids_by_run[run_str]['pulse_ids']
            self.train_ids = ids_by_run[run_str]['train_ids']
            self.baseline_shifts = ids_by_run[run_str]['baseline_shift']
            self.frame_mask = ids_by_run[run_str]['frame_mask']
        else:
            load = self.db.load
            pm = {'run':run,'data_mode':self.data_mode}
            cell_ids = load('cell_ids',path_modifiers = pm)
            pulse_ids = load('pulse_ids',path_modifiers = pm)
            train_ids = load('train_ids',path_modifiers = pm)
            frame_mask = load('frame_mask',path_modifiers = pm)
            baseline_shifts= load('baseline_shift',path_modifiers = pm)
            self.ids_by_run[run_str] = {'cell_ids':cell_ids,'pulse_ids':pulse_ids, 'train_ids':train_ids,'frame_mask':frame_mask,'baseline_shift':baseline_shifts}
            self.cell_ids = cell_ids
            self.pulse_ids = pulse_ids
            self.train_ids = train_ids
            self.frame_mask = frame_mask
            self.baseline_shifts = baseline_shifts
        self.nframes=len(self.cell_ids)
        
    
    ## setup filters to be applied to the data stream ##
    def load_roi_manager(self):
        used_roi_names = self.collect_used_roi_names()
        roi_manager = ROIManager(self.get_geometry(),used_rois = used_roi_names,rois_dict = self.opt['ROIs'])
        return roi_manager
    def collect_used_roi_names(self):
        filter_opt = self.opt['filters']
        roi_names = []
        for s_name in self.opt['filter_sequence']:
            try:
                roi_names += filter_opt[s_name].get('ROIs',['all'])
            except KeyError as e:
                log.warning('Quantity {} does not exist.Skipping.'.format(name))
        used_roi_names = list(np.unique(roi_names))
        return used_roi_names 
    
    def load_filters(self):
        filter_sequence = []
        for filter_name in self.opt.get('filter_sequence',[]):
            filter_opt = self.opt['filters'][filter_name]
            filter_opt['name']=filter_name
            roi_mask = self.roi_manager.get_combined_roi_mask(filter_opt.get('ROIs',['all']))
            filter_opt['roi_mask'] = roi_mask
            has_unmasked_values = roi_mask.any()
            if not has_unmasked_values:
                log.warning('The roi of filter {} has no unmasked values. Skipping.'.format(filter_name))
                continue
            class_name = filter_opt.get('class')
            try:
                filter_class = getattr(filters,class_name)
                #log.info('Creating filter from class {}'.format(filter_class))
            except AttributeError as e:
                log.warning('Filter {} not found in {}, skipping this Filter'.format(class_name,filters))
            filter_sequence.append(filter_class(filter_opt))
        filter_sequence = FilterSequence(filter_sequence)
        return filter_sequence            

    
    ## load data from database and apply filters ##
    def _selection_to_chunks(self,opt:dict):
        ex_opt = settings.experiment
        run = opt['run']
        if not str(run) in self.info:
            # The call to load info potentially creates a new vds file if non is existing.
            self.info[str(run)] = self.db.load('info',runs=[run],data_mode=self.data_mode)[str(run)]
            
        modules =np.array(opt['modules'],dtype = int)
        frame_range = opt['frame_range']
        selection = {key:Selection(key,item) for key,item in opt['selection'].items()}
        good_cells = self.opt['good_cells']
        #calib=self.calibrator
        #log.info('calibrator wants to set ids.')

        ### self set ids is iportant otherwise metadata will be wrong
        self.set_ids_for_run(run)
        
        train_selection = selection['trains']
        min_train_id = np.min(self.train_ids)
        n_trains = len(self.train_ids)
        
        
        #log.info(f'train ids min = {np.max(self.train_ids)} max = {np.max(self.train_ids)}')
        log.info('selection = {}'.format({s.name:s.data_range for s in selection.values()}))
        comm = self.comm_module
        
        chunk_opt = {
            'run':run,
            'data_mode':self.data_mode,
            'n_cpus':comm.n_cpus(),
            'free_mem':comm.free_mem()*self.opt['RAM_multiplier'],
            'selection': selection,
            'n_frames': opt['n_frames'],
            'frame_range': frame_range,
            'good_cells': good_cells,
            'in_multiples_of': opt['in_multiples_of'],
            'modules' : opt['modules']
        }
        db_load = self.db.load        
        chunks = db_load('chunks',**chunk_opt)
        if ex_opt.get('load_pump_diod_data',False):
            diod_dict = self.db.load('pump_diod',path_modifiers={'run':run})
            diod_on = diod_dict['diod_on'].astype(bool)
            diod_trains = diod_dict['trainIds']
            min_train_id = self.train_ids.min()
        #log.info(f'chunks = {chunks}')
        chunked_metadata=[]
        for chunk in chunks:
            chunk_train_ids = self.train_ids[chunk]
            _dict = {'pulse_ids':self.pulse_ids[chunk],'cell_ids':self.cell_ids[chunk],'train_ids':chunk_train_ids,'good_frames_mask':self.frame_mask[chunk],'baseline_shift':self.baseline_shifts[chunk]}
            if ex_opt.get('load_pump_diod_data',False):
                if isinstance(diod_on,np.ndarray):
                    _dict['diod_on'] = diod_on[chunk_train_ids-min_train_id]
                    _dict['diod_trains'] = diod_trains[chunk_train_ids-min_train_id]
            chunked_metadata.append(_dict)            
        return chunks,chunked_metadata

    def _process_data_chunk_worker(self,modules,slices,run,data_slices,output_slices,data_mode,filter_sequence,custom_mask,frame_mask,module_id_lookup,**opt):
        data,mask,gain,filtered_mask = opt['outputs']
        #process_events = opt['events']
        process_id = opt['local_name']
        n_processes = opt['number_of_processes']
        synchronize = opt['synchronize']

        # let database routin populate the outarrays
        self.db._load_data_chunk_worker(modules,slices,module_id_lookup,run,data_slices,output_slices,data_mode,**opt)
        #process_events[process_id][0].set()
        
        #wait until all other workers finished loading the data into the shared outputs
        #for p_id,event in enumerate(process_events):
        #    if p_id == process_id:
        #        continue
        #    success = event[0].wait(timeout = 60*20)
        #    if not success:
        #        log.warning('Process {} didnt synchronice after 20 min stop waiting in process {}'.format(p_id,process_id))
        synchronize(timeout = 60*20)
        #log.info('all processes synchronized')
        
        #take this workers pice of data: assumes module,frames,long dim,short dim layout
        n_frames = data.shape[1]
        selection = slice(process_id,None,n_processes)
        #log.info('selection = {}'.format(selection))
        data_view = np.swapaxes(data[:,selection],0,1)
        mask_view = np.swapaxes(mask[:,selection],0,1)        
        mask_view[:] &= custom_mask[None,:]
        gain_view = np.swapaxes(gain[:,selection],0,1)
        # Note that all variables called *_view are really numpy views into the big shared arrays data,mask,gain.
        # If one edits their elements the elements of the shared arrays change.
        views = {'data':data_view,'mask':mask_view,'gain':gain_view}
        init_filtered_mask = ~frame_mask[selection]
        #log.info('filtered_mask shape = {}'.format(init_filtered_mask.shape))
        init_modified_mask = np.zeros_like(init_filtered_mask)
        
        
        # The following is bad if not all workers are past the synchronisation phase yet.S        
        
        views,masks = filter_sequence(views,masks=FilterTools.init_masks(filtered_mask = init_filtered_mask,modified_mask = init_modified_mask))
        #log.info('filtered_mask shape after sequence = {}'.format(masks['total_filtered'].shape))
        filtered_mask[selection] = masks['total_filtered']
        #process_events[process_id][1].set()
        #wait until all other workers finished generating the filter_masks
        #for p_id,event in enumerate(process_events):
        #    if p_id == process_id:
        #        continue
        #    success = event[1].wait(timeout = 60*20)
        #    process_events[process_id][0].clear()
        #    if not success:
        #        log.warning('Process {} didnt synchronice after 20 min stop waiting in process {}'.format(p_id,process_id))
        synchronize(timeout = 60*20)

    def _process_data_chunk_worker2(self,modules,run,data_slices,output_slices,data_mode,filter_sequence,custom_mask,frame_mask,module_id_lookup,**opt):
        data,mask,gain,filtered_mask = opt['outputs']
        #process_events = opt['events']
        process_id = opt['local_name']
        n_processes = opt['number_of_processes']
        synchronize = opt['synchronize']

        
        # let database routin populate the outarrays
        for module in modules:
            self.db._load_data_chunk_worker2(module,module_id_lookup,run,data_slices,output_slices,data_mode,**opt)
        #process_events[process_id][0].set()
        
        #wait until all other workers finished loading the data into the shared outputs
        #for p_id,event in enumerate(process_events):
        #    if p_id == process_id:
        #        continue
        #    success = event[0].wait(timeout = 60*20)
        #    if not success:
        #        log.warning('Process {} didnt synchronice after 20 min stop waiting in process {}'.format(p_id,process_id))
        synchronize(timeout = 60*20)
        #log.info('all processes synchronized')
        
        #take this workers pice of data: assumes module,frames,long dim,short dim layout
        n_frames = data.shape[1]
        selection = slice(process_id,None,n_processes)
        #log.info('selection = {}'.format(selection))
        data_view = np.swapaxes(data[:,selection],0,1)
        mask_view = np.swapaxes(mask[:,selection],0,1)        
        mask_view[:] &= custom_mask[None,:]
        gain_view = np.swapaxes(gain[:,selection],0,1)
        # Note that all variables called *_view are really numpy views into the big shared arrays data,mask,gain.
        # If one edits their elements the elements of the shared arrays change.
        views = {'data':data_view,'mask':mask_view,'gain':gain_view}
        init_filtered_mask = ~frame_mask[selection]
        #log.info('filtered_mask shape = {}'.format(init_filtered_mask.shape))
        init_modified_mask = np.zeros_like(init_filtered_mask)
        
        
        # The following is bad if not all workers are past the synchronisation phase yet.S        
        
        views,masks = filter_sequence(views,masks=FilterTools.init_masks(filtered_mask = init_filtered_mask,modified_mask = init_modified_mask))
        #log.info('filtered_mask shape after sequence = {}'.format(masks['total_filtered'].shape))
        filtered_mask[selection] = masks['total_filtered']
        #process_events[process_id][1].set()
        #wait until all other workers finished generating the filter_masks
        #for p_id,event in enumerate(process_events):
        #    if p_id == process_id:
        #        continue
        #    success = event[1].wait(timeout = 60*20)
        #    process_events[process_id][0].clear()
        #    if not success:
        #        log.warning('Process {} didnt synchronice after 20 min stop waiting in process {}'.format(p_id,process_id))
        synchronize(timeout = 60*20)
                
    def get_data(self,opt:AgipdDataSelection):
        ex_opt = settings.experiment
        run = opt['run']
        if not str(run) in self.info:
            # The call to load info potentially creates a new vds file if non is existing.
            self.info[str(run)] = self.db.load('info',runs=[run],data_mode=self.data_mode)[str(run)]
            
        modules =np.array(opt['modules'],dtype = int)
        module_id_lookup = {m:m_id for m_id,m in enumerate(modules)} 
        frame_range = opt['frame_range']
        good_cells = self.opt['good_cells']
        
        self.roi_manager.used_modules = np.asarray(modules)
        #log.info('modules = {}'.format(modules))
        nroi = self.roi_manager.rois['normalization_region']
        #log.info('normalization mask ={} complete mask = {}'.format(nroi.mask.shape,nroi.mask_complete.shape))
        
        filter_sequence = self.load_filters()
        apply_filter_sequence = filter_sequence.apply
        chunks,chunked_metadata = self._selection_to_chunks(opt)
            
        comm = self.comm_module        
        n_frames_to_process=np.sum(tuple(chunk.shape[0] for chunk in chunks))        
        custom_mask = self.custom_mask[modules]
        
        data_shape=self.info[str(run)]['data_shape']        
        n_processes = ex_opt.n_processes
        
        if self.opt['data_mode'] == 'proc':
            def mp_process_data(c_id,chunk):                
                frame_mask = chunked_metadata[c_id]['good_frames_mask']
                xprint('Loading data chunk {} of {} with {} patterns'.format(c_id+1,len(chunks),len(chunk)))                
                chunk_shape = (len(modules),len(chunk))+data_shape[1:]
                len_chunk = len(chunk)
                frame_slices,out_slices = split_into_simple_slices(chunk,return_sliced_args=True)
                #log.info(frame_slices)
                slice_length = [(s.stop-s.start) for s in frame_slices]
                #log.info('mean slice_length = {} std = {}'.format(np.mean(slice_length),np.std(slice_length)))
                
                output_shapes =  [chunk_shape,chunk_shape,chunk_shape,(len_chunk,)]
                output_dtypes = [np.dtype('float32'),np.dtype('bool'),np.dtype('uint8'),np.dtype('bool')] 
                mp_mode = Multiprocessing.MPMode_SharedArray(output_shapes,output_dtypes)
                data,mask,gain,filtered_mask= self.comm_module.request_mp_evaluation(self._process_data_chunk_worker,mp_mode,input_arrays = [modules,np.arange(len(frame_slices))], const_inputs = [run,frame_slices,out_slices,self.data_mode,apply_filter_sequence,custom_mask,frame_mask,module_id_lookup], call_with_multiple_arguments = True, split_mode = 'modulus', n_processes = n_processes)
                #data,mask,gain,filtered_mask= self.comm_module.request_mp_evaluation(self._process_data_chunk_worker2,mp_mode,input_arrays = [modules], const_inputs = [run,frame_slices,out_slices,self.data_mode,apply_filter_sequence,custom_mask,frame_mask,module_id_lookup], call_with_multiple_arguments = True, split_mode = 'modulus', n_processes = n_processes)                
                data = data.swapaxes(0,1)
                mask = mask.swapaxes(0,1)
                gain = gain.swapaxes(0,1)
                out_dict =  {'data':data,'mask':mask,'gain':gain,'frame_ids':chunk,'filtered_frames_mask':filtered_mask,**chunked_metadata[c_id]}
                #log.info('removing filtered_elements')
                #out_dict = FilterTools.remove_filtered_elements(out_dict,filtered_mask)
                #log.info('ex worker mask containes {} % unmasked values'.format(np.sum(mask_chunk)/np.prod(mask_chunk.shape)))
                #log.info('datatype data_chunk = {}'.format(raw_chunk[0].dtype))
                #log.info('datatype mask_chunk = {}'.format(mask_chunk[0].dtype))
                return out_dict

        else:
            raise NotImplementedError('Currently only loading from processed data is supported. (data_mode has to be "proc")')
        
        for c_id,chunk in enumerate(chunks):
            yield mp_process_data(c_id,chunk)



    ## setup detector geometry in reciprocal space ##
    ## i.e. the reciprocal space coordinates associated with each pixel id the detector ##
    def get_pixel_grid_reciprocal(self,approximation='None',out_coord_sys='spherical'):
        grid = self.detector.pixel_grid
        pixel_grid = pLib.pixel_grid_to_scattering_grid(grid,self.x_ray_wavelength,approximation = approximation, out_coord_sys = out_coord_sys)
        return pixel_grid

    def _get_geometry_rois(self, approximation='None'):
        mask = self.detector.sensitive_pixel_mask
        data_shape = self.detector.data_shape
        grid = self.detector.pixel_grid[:,:-1,:-1][mask].reshape(data_shape+(3,))        
        data_grid_spher = pLib.pixel_grid_to_scattering_grid(grid,self.x_ray_wavelength,approximation = approximation, out_coord_sys = 'spherical')
        return {'data_grid_spherical':data_grid_spher,'data_shape':data_shape,'asic_slices':self.detector.asic_slices}

    def get_geometry(self, approximation='None',out_coord_sys='spherical'):
        grid = self.detector.pixel_grid
        framed_grid = self.detector.framed_pixel_grid
        framed_centers = self.detector.framed_pixel_centers
        pixel_grid = pLib.pixel_grid_to_scattering_grid(grid,self.x_ray_wavelength,approximation = approximation, out_coord_sys = out_coord_sys)
        framed_pixel_grid = pLib.pixel_grid_to_scattering_grid(framed_grid,self.x_ray_wavelength,approximation = approximation, out_coord_sys = out_coord_sys)
        framed_pixel_centers = pLib.pixel_grid_to_scattering_grid(framed_centers,self.x_ray_wavelength,approximation = approximation, out_coord_sys = out_coord_sys)

        mask = self.detector.sensitive_pixel_mask
        data_shape = self.detector.data_shape
        grid = self.detector.pixel_grid[:,:-1,:-1][mask].reshape(data_shape+(3,))        
        data_grid_spher = pLib.pixel_grid_to_scattering_grid(grid,self.x_ray_wavelength,approximation = approximation, out_coord_sys = 'spherical')
        
        return {'pixel_grid':pixel_grid,'lab_pixel_grid':grid,'framed_pixel_grid':framed_pixel_grid,'framed_lab_pixel_grid':framed_grid,'framed_pixel_centers':framed_pixel_centers,'framed_lab_pixel_centers':framed_centers,'mask':mask,'framed_mask':self.detector.framed_sensitive_pixel_mask,'data_shape':data_shape,'asic_slices':self.detector.asic_slices,'data_grid_spherical':data_grid_spher}
    

    ## satisfy interface ##
    ## currently empty since experiment and analysis workers are not disconnected in separate processes yet##
    def run(self):
        pass           



import numpy as np
import os
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1" 
os.environ['OPENBLAS_NUM_THREADS'] = "1"

os.chdir('/gpfs/exfel/theory_group/user/berberic/Programs/xframe')
from xframe.database.euxfel import default_DB
from xframe.library.gridLibrary import GridFactory
from xframe.startup_routines import dependency_injection
dependency_injection()
from xframe.presenters.matplolibPresenter import plot1D,heat2D_multi,heat2D,heatPolar2D,agipd_heat
from xframe.library.mathLibrary import circularHarmonicTransform_real_forward as ht
from xframe.library.mathLibrary import cartesian_to_spherical,spherical_to_cartesian,SampleShapeFunctions
from xframe.library.mathLibrary import circularHarmonicTransform_real_inverse as iht

from xframe.plugins.fxs3046_online.analysisLibrary.regrid2 import SimpleRegridder,create_lookup_array_2D
from xframe.plugins.fxs3046_online.analysisLibrary.cross_correlation import generate_calc_cross_correlation

from scipy.spatial.kdtree import KDTree
from scipy.ndimage import gaussian_filter
from scipy.ndimage import map_coordinates
import timeit
import time


db = default_DB()



def interpolate(values, vtx, wts):
    return np.einsum('nj,nj->n', np.take(values, vtx), wts)
#
#from xframe.startup_routines import load_recipes
#from xframe.control.Control import Controller
#analysis_sketch,exp_sketch = load_recipes(ana_name='fxs3046_online.generate_plots_2', ana_settings='hist_mean_162',exp_name='fxs3046_online',exp_settings = 'proc_-20_20_03_4')
#controller = Controller(analysis_sketch,experiment_sketch = exp_sketch)
#controller.chooseJobSingleProcess()
#from xframe import database
#from xframe import Multiprocessing
#
#comm = Multiprocessing.comm_module
#g = comm.get_geometry()

base_path = '/gpfs/exfel/theory_group/user/berberic/fxs3046/test/ccn/'

#fig=agipd_heat.get_fig(data,grid_cart,mask)
#db.save(base_path + 'data.matplotlib',fig)

#xs = np.arange(-5.5,6,1)
#pixel_centers = GridFactory.construct_grid('uniform',[xs,xs])[:]
#pixel_centers_polar = cartesian_to_spherical(pixel_centers)
#data = np.zeros(pixel_centers.shape[:-1])
#data[5:7]=1
poly_func = SampleShapeFunctions.get_polygon_function(2,6,coordSys = 'polar')
poly_func_cart = SampleShapeFunctions.get_polygon_function(2,6,coordSys = 'cartesian')


centers = spherical_to_cartesian(np.array([(0,0)] + [(1,phi*2*np.pi/5) for phi in range(5) ]))
densities = [25,50,25,50,25,50]

#for n,i in enumerate(data[1:-1]):
#    i[1+n%2:-1:2]=1


#fig = heat2D.get_fig(data,grid = pixel_centers)

#bound = np.min(np.max(spherical_to_cartesian(pixel_centers),axis = (0,1,2))[:2])
bound = 12
n_r_steps = 512
rs = np.arange(n_r_steps//2)*2*bound/n_r_steps
xs = np.arange(-n_r_steps//2,n_r_steps//2)*2*bound/n_r_steps

xs2 = np.arange(-n_r_steps,n_r_steps)*2*bound/n_r_steps

irs = np.pi/(bound)*np.arange(n_r_steps//2)
ixs = np.pi/bound*np.arange(-n_r_steps//2,n_r_steps//2)
ixs2 = np.pi/bound*np.arange(-n_r_steps,n_r_steps)/2


n_phis = 1024
phis = np.arange(n_phis)*2*np.pi/n_phis
cart_grid = GridFactory.construct_grid('uniform',[xs,xs])[:]
polar_grid = GridFactory.construct_grid('uniform',[rs,phis])[:]

icart_grid2 = GridFactory.construct_grid('uniform',[ixs2,ixs2])[:]
icart_grid = GridFactory.construct_grid('uniform',[ixs,ixs])[:]
ipolar_grid = GridFactory.construct_grid('uniform',[irs,phis])[:]

polar_data = poly_func(polar_grid)

#cart_data = poly_func_cart(cart_grid)
cart_data = SampleShapeFunctions.get_disk_function(0.5,lambda points:np.full(points.shape[:-1],densities[0]),center=centers[0],norm='standard',coordSys='cartesian')(cart_grid)
for c,d in zip(centers[1:],densities[1:]):
    cart_data+=SampleShapeFunctions.get_disk_function(0.5,lambda points:np.full(points.shape[:-1],d),center=c,norm='standard',coordSys='cartesian')(cart_grid)
    
pixel_size = ixs[1] - ixs[0]
n_pixels = 2*np.pi*irs//pixel_size
n_orders = (n_pixels//2 + 1 ).astype(int)

cc = generate_calc_cross_correlation(polar_data.shape,n_orders,mask_type='unmasked')
rg = SimpleRegridder(cartesian_to_spherical(icart_grid),ipolar_grid)
rg2 = SimpleRegridder(icart_grid,icart_grid2,coord_sys='cartesian',interpolation = 'linear')
rg3 = SimpleRegridder(icart_grid,icart_grid2,coord_sys='cartesian',interpolation = 'nearest')
rgl = SimpleRegridder(cartesian_to_spherical(icart_grid),ipolar_grid,interpolation='linear')

#polar_data = rg.apply(data)
ftd = np.fft.fft2(cart_data)
I = (ftd*ftd.conj()).real
I = np.roll(I,n_r_steps//2,axis=(0,1))


polar_I=rg.apply(I)
l_polar_I = rgl.apply(I)
#print(timeit.timeit('fu(I)',number=100,globals = globals()))

spline_polar_grid = spherical_to_cartesian(GridFactory.construct_grid('uniform',[np.arange(n_r_steps//2)*2,phis])[:]) + np.array(I.shape)-0.5
spline_polar_grid2 = spherical_to_cartesian(GridFactory.construct_grid('uniform',[np.arange(n_r_steps//2),phis])[:]) + np.array(I.shape)/2-0.5
s_polar_I2 = map_coordinates(rg2.apply(I),(spline_polar_grid[...,0],spline_polar_grid[...,1]))
s_polar_I3 = map_coordinates(rg3.apply(I),(spline_polar_grid[...,0],spline_polar_grid[...,1]))
s_polar_I = map_coordinates(I,(spline_polar_grid2[...,0],spline_polar_grid2[...,1])) 
#print(timeit.timeit('map_coordinates(I,(spline_polar_grid[...,0],spline_polar_grid[...,1]))',number=100,globals = globals()))

cc_data = cc(polar_I)
start = time.time()
Im = np.swapaxes(ht(s_polar_I),0,1)
lIm = np.swapaxes(ht(s_polar_I2),0,1)
nIm = np.swapaxes(ht(s_polar_I3),0,1)
llIm = np.swapaxes(ht(l_polar_I),0,1)

#for i in range(n_r_steps//2):
#    Im[n_orders[i]:,i]=0
#
ccm = np.abs(Im[:,:,None]*Im.conj()[:,None,:]) #np.moveaxis(np.abs(ht(cc_data)),2,0)
lccm = np.abs(lIm[:,:,None]*lIm.conj()[:,None,:]) #np.moveaxis(np.abs(ht(cc_data)),2,0)
nccm = np.abs(nIm[:,:,None]*nIm.conj()[:,None,:]) #np.moveaxis(np.abs(ht(cc_data)),2,0)
llccm = np.abs(llIm[:,:,None]*llIm.conj()[:,None,:]) #np.moveaxis(np.abs(ht(cc_data)),2,0)
print(time.time()-start)
ccm_grid = GridFactory.construct_grid('uniform',[irs,irs])[:]
#
#db.save(base_path + 'polar_data.vts',[polar_data],grid = polar_grid,grid_type='polar')
#db.save(base_path + 'cart_data.vtr',[cart_data],grid = cart_grid,grid_type='cartesian')
#db.save(base_path + 'intensity.vtr',[I],grid = cart_grid,grid_type='cartesian')
db.save(base_path + 'polar_I.vts',[polar_I,l_polar_I,s_polar_I],dset_names=['NN','linear','spline'],grid = ipolar_grid,grid_type='polar')

orders = range(0,30,2)
db.save(base_path + 'ccm.vtr',[ccm[i] for i in orders],dset_names = ['C{}'.format(o) for o in orders],grid = ccm_grid,grid_type='cartesian')
db.save(base_path + 'lccm.vtr',[lccm[i] for i in orders],dset_names = ['lC{}'.format(o) for o in orders],grid = ccm_grid,grid_type='cartesian')
db.save(base_path + 'llccm.vtr',[llccm[i] for i in orders],dset_names = ['llC{}'.format(o) for o in orders],grid = ccm_grid,grid_type='cartesian')
db.save(base_path + 'nccm.vtr',[lccm[i] for i in orders],dset_names = ['nC{}'.format(o) for o in orders],grid = ccm_grid,grid_type='cartesian')

#options = {'plot_range':[1e7,1e16]}
#def _save_first_invariants(bls,radial_points,base_path,options,name='model_0_'):
#    max_value = np.max(tuple(np.abs(bl).max() for bl in bls))
#    vmin,vmax = options.get('plot_range',[max_value*1e-12,max_value])
#    if [vmin,vmax] == [None,None]:
#        vmin,vmax = [max_value*1e-12,max_value]
#        
#    grid = GridFactory.construct_grid('uniform',[radial_points,np.arange(bls.shape[2])])
#    order_ids = [[2,4,16,24,34],[36,38,40,42,44]]
#    #order_ids = [20,40,60,80,127]
#    
#    shape = [2,len(order_ids[0])]
#    layouts = []
#    for i in range(shape[0]):
#        layout_part = []
#        orders = np.array(order_ids[i])
#        #orders = np.arange(10*i,10*i+len(order_ids)*2,2)
#        for o in orders:
#            layout = {'title':'$C_{'+'{}'.format(o)+'}$',
#                      'x_label':'$q_1$',
#                      'y_label':'$q_2$'
#                      }
#            layout_part.append(layout)
#        layouts.append(layout_part)
#                                                
#    fig_bl_masks = heat2D_multi.get_fig([[np.abs(bls[n]) for n in order_ids[0]],[np.abs(bls[n]) for n in order_ids[1]] ],scale = 'log',layout = layouts,grid =grid[:],shape = shape,size = (30,10),vmin= vmin, vmax = vmax,cmap='plasma')
#    bl_path = base_path +name+'ccn.matplotlib'
#    db.save(bl_path,fig_bl_masks,dpi = 300)
#    
##_save_first_invariants(ccm,irs,base_path,options,'pentagon_')



import abc
from xframe.interfaces import DependencyMeta
class DatabaseInterface(abc.ABC):
    @abc.abstractmethod
    def save(self,name,data,**kwargs):
        pass
    
    @abc.abstractmethod
    def load(self,name,**kwargs):
        pass


class SphericalHarmonicTransformInterface(abc.ABC):
    '''
    Interface for the external library shtns which supplies spherical harmonic transforms
    '''
    _external_dependency_name_ = 'shtns'
    @abc.abstractmethod
    def phi(self):
        pass
    @abc.abstractmethod
    def theta(self):
        pass
    @abc.abstractmethod
    def forward_l(self):
        pass
    @abc.abstractmethod
    def forward_m(self):
        pass
    @abc.abstractmethod
    def inverse_m(self):
        pass
    @abc.abstractmethod
    def inverse_l(self):
        pass

class SphericalHarmonicTransformDependency(SphericalHarmonicTransformInterface,metaclass=DependencyMeta):
    pass

class SoftInterface(abc.ABC):
    '''
    Interface for the external library shtns which supplies spherical harmonic transforms
    '''
    _external_dependency_name_ = 'pysofft'
    @abc.abstractmethod
    def forward_cmplx(self):
        pass
    @abc.abstractmethod
    def inverse_cmplx(self):
        pass

class SoftDependency(SoftInterface,metaclass=DependencyMeta):
    pass
    
class DiscreteLegendreTransform_interface(abc.ABC):
    '''
    Interface for the external library shtns which supplies spherical harmonic transforms
    '''
    _external_dependency_name_ = 'flt'
    @abc.abstractmethod
    def forward():
        pass
    @abc.abstractmethod
    def inverse():
        pass
    
class DiscreteLegendreTransformDependency(DiscreteLegendreTransform_interface,metaclass=DependencyMeta):
    pass

class GSLInterface(abc.ABC):
    '''
    Interface for the Gnu Scientific Library wraper 
    '''
    _external_dependency_name_ = 'gsl'
    @abc.abstractmethod
    def legendre_sphPlm_array(l_max,m_max,xs,return_orders = False,sorted_by_l = False):
        pass
    @abc.abstractmethod
    def bessel_jl(ls,xs):
        pass
    @abc.abstractmethod
    def hyperg_2F1(a,b,c,z):
        pass    
class GSLDependency(GSLInterface,metaclass=DependencyMeta):
    pass
    
class PeakDetectorInterface(abc.ABC):
    '''
    Interface for the Gnu Scientific Library wraper 
    '''
    _external_dependency_name_ = 'persistent_homology'
    @abc.abstractmethod
    def find_peaks(dim,data):
        raise AssertionError('PeakDetector was not jet dependency injected.')

class PeakDetectorDependency(PeakDetectorInterface,metaclass=DependencyMeta):
    pass



import tempfile
import os
import numpy as np

import logging
import pdb_eda
from xframe.library.gridLibrary import GridFactory
from xframe.library.pythonLibrary import DictNamespace
from xframe.database.interfaces import PDBInterface
from xframe.library.mathLibrary import spherical_to_cartesian,gaussian_fft_cart,get_fft_reciprocal_grid,convolve_with_gaussian_cart,SphericalIntegrator
log=logging.getLogger('root')

class ProteinDB(PDBInterface):
    lib = pdb_eda
    @staticmethod
    def load(pdb_id : str,**kwargs):
        '''
        Loads density and grid of a pdb entry. Needs internet connection to work. 
        throws assertion error if no density is accessible for given pdb entry or if id does not exist.
        :param pdb_id: Proteine Id on https://www.rcsb.org/ e.g.: '1KP8' for GroEL 
        :type str: 
        :return density_and_grid: 
        :rtype DictNamespace: keys()=('density','grid')
        '''
        return DensityExtractor(pdb_id)
        
    @staticmethod
    def save(path,data,**kwargs):
        e=NotImplementedError('Saving to PDB not possible.')
        raise e

class DensityExtractor:
    def __init__(self,pdb_id):
        cwd = os.getcwd()
        with tempfile.TemporaryDirectory() as tmpdirname:
            os.chdir(tmpdirname)
            log.info('Start loading files from PDB. This might take a minute..')
            densityAnalysisObj = pdb_eda.densityAnalysis.fromPDBid(pdb_id,downloadFile = False,pdbi=True,mmcif=False,pdbbio=True)
            assert densityAnalysisObj != 0, 'given pdb_id = {} dies not exist or does not contain 2fo-fc density data'.format(pdb_id)
            self._densities= DictNamespace(
                fo = densityAnalysisObj.fo.density.T,
                fc = densityAnalysisObj.fc.density.T,
                two_fofc = densityAnalysisObj.densityObj.density.T,
                fofc = densityAnalysisObj.diffDensityObj.density.T
            )
            h=densityAnalysisObj.densityObj.header
            self.header_part = DictNamespace(
                n_crs = np.asarray(h.ncrs),
                grid_length = np.asarray(h.gridLength),
                origin = np.asarray(h.origin),
                xyz_interval = np.asarray(h.xyzInterval),
                crs_start = np.asarray(h.crsStart),
                map_to_xyz = np.asarray(h.map2xyz),
                map_to_crs = np.asarray(h.map2crs),
                ortho_mat = np.asarray(h.orthoMat),
                inv_ortho_mat = np.asarray(h.deOrthoMat),
                alpha = h.alpha,
                beta = h.beta,
                gamma = h.gamma        
            )
            self._atom_coords_xyz = np.asarray(densityAnalysisObj.asymmetryAtomCoords)

        self.atom_coords = DictNamespace(
            xyz = self._atom_coords_xyz,
            crs = self.xyz2crsCoord(self._atom_coords_xyz)
        )
        self.bounds = DictNamespace(
            xyz = np.array(tuple((self.atom_coords.xyz[...,n].min(),self.atom_coords.xyz[...,n].max()) for n in range(3))),
            crs = np.array(tuple((self.atom_coords.crs[...,n].min(),self.atom_coords.crs[...,n].max()) for n in range(3)))            
        )
        self.max_r = np.linalg.norm(self.atom_coords.xyz,axis= -1).max()
        os.chdir(cwd)
                
    def crs2xyzCoord(self,crs):
        '''
        Convert the crs coordinates into xyz coordinates.
        Assumes crs is of shape (N,3) where N is the number of coordinate sets.
        '''
        h = self.header_part
        crs = np.atleast_2d(crs)
        if h.alpha == h.beta == h.gamma == 90:
            xyz = crs[:,h.map_to_xyz]*h.grid_length[None,:]+h.origin[None,:]
            
        else:
            xyz = np.dot(h.ortho_mat, ((crs[:,h.map_to_xyz]+h.crs_start[None,h.map_to_xyz])/h.xyz_interval[None,:]).T).T
        return np.squeeze(xyz)
    
    def xyz2crsCoord(self,xyz):
        '''
        Converts xyz to crs coordinates. 
        Assumes xyz is of shape (N,3) where N is the number of coordinate sets.
        '''
        h = self.header_part
        xyz = np.atleast_2d(xyz)
        if h.alpha == h.beta == h.gamma == 90:
            crsGridPos =  np.rint((xyz - h.origin[None,:])/h.grid_length[None,:])[:,h.map_to_crs]
        else:
            fraction = np.dot(h.inv_ortho_mat, xyz.T).T
            crsGridPos = (np.rint(fraction*h.xyz_interval[None,:]) - h.crs_start[None,h.map_to_xyz])[:,h.map_to_crs]
        return np.squeeze(crsGridPos).astype(int)

    def sample_density_xyz(self,crs_density,points,fill_value = 0.0):
        '''
        Samples density at given xyz coordinates.
        '''
        shape = points.shape
        crs_points = self.xyz2crsCoord(points.reshape(-1,3))
        sampled_density = self.sample_density_crs(crs_density,crs_points,fill_value = fill_value).reshape(shape[:-1])
        return sampled_density
    
    def sample_density_crs(self,crs_density,points,fill_value = 0.0):
        '''
        Samples density at given crs coordinates.
        fills defaut value for crs coordinates which are smaller than 0 and bigger than self.n_crs.
        '''
        points = np.atleast_2d(points)
        shape = points.shape
        log.info('shape crs = {}'.format(shape))
        valid_crs_points , mask = self.return_valid_crs_coords(points.reshape(-1,3),return_mask= True)

        sampled_density=np.full(points.shape[:-1],fill_value,dtype = crs_density.dtype)
        log.info('sample density shape = {}'.format(sampled_density.shape))
        sampled_density[mask] = crs_density[valid_crs_points[:,0],valid_crs_points[:,1],valid_crs_points[:,2]]
        return sampled_density.reshape(shape[:-1])
    def return_valid_crs_coords(self,coords,return_mask = False):
        '''
        Returns all coordinates tripples that satisfy  0<= val[i] < heade.n_crs[i] for i in 0,1,2.
        '''
        valid_mask = (np.prod( (coords < self.header_part.n_crs[None,:]) * (coords >= 0),axis=-1)).astype(bool)
        coords = coords[valid_mask,:]
        if return_mask:
            return coords,valid_mask
        else:
            return coords
        
    def generate_density_mask(self,atom_neighbour_shells=2):
        '''
        Generates a mask around each atom of the PDB structure. (To get rid of symmetry parts of the density fuction.)
        :param atom_neighbor_shells: Number of cubical shells around a given crs coordinate which to include in the density
        :type int: 
        '''
        mask = np.full(self.header_part.n_crs,False)
        if atom_neighbour_shells >=1:
            crs = self.atom_coords.crs
            neighbours_1d = np.arange(-atom_neighbour_shells ,atom_neighbour_shells + 1)
            neighbours_3d = GridFactory.construct_grid('uniform',[neighbours_1d,neighbours_1d,neighbours_1d]).flatten().array
            valid_grid_positions = np.unique(self.return_valid_crs_coords((crs[:,None,:] + neighbours_3d[None,...]).reshape(-1,3)),axis = 0)        
            mask[valid_grid_positions.T[0],valid_grid_positions.T[1],valid_grid_positions.T[2]]=True
        else:
            mask[:]=True
        return mask

    def convolve_density_with_gaussian(self,crs_density,sigma):
        '''
        Convolves a density in self._densities with a gaussian.
        '''
        n_crs = self.header_part.n_crs        
        n_steps = 2**(np.log2(2*n_crs.max()) + 0.5).astype(int)
        grid_bounds = np.max(np.abs(self.bounds.xyz),axis=-1)
        step_size = grid_bounds/(n_steps//2)
        
        steps = [np.arange(-n_steps//2+1,n_steps//2+1) * step_size[i] for i in range(3)]        
        cart_grid = GridFactory.construct_grid('uniform',steps).array
        density_cart = self.sample_density_xyz(crs_density,cart_grid)
        
        convolved_density_cart = convolve_with_gaussian_cart(density_cart,cart_grid,sigma)
        log.info('convolved .shape = {}'.format(convolved_density_cart.shape))

        crs_grid = GridFactory.construct_grid('uniform',[np.arange(n) for n in n_crs]).array
        xyz_grid = self.crs2xyzCoord(crs_grid.reshape(-1,3)).reshape(crs_grid.shape)
        mask = np.prod((xyz_grid >= -grid_bounds[None,None,None,:]) & (xyz_grid <= grid_bounds[None,None,None,:]),axis = -1).astype(bool)
       # log.info('mas is true = {}'.format(mask.all()))
        #log.info("{} masked ccordinates".format(sum(~mask.flatten())))
        nearest_neighbor_indices = ((xyz_grid+grid_bounds[None,None,None,:])//step_size[None,None,None,:]).astype(int)
        #log.info('{} indices bigger than 128!'.format(np.sum(nearest_neighbor_indices>=128)))
        #log.info('mask shape = {}'.format(mask.shape))
        nni = nearest_neighbor_indices[mask]
        #log.info('{} indices bigger than 128!'.format(np.sum(nni>=128)))
        #log.info('nni shape unmasked = {}, nni shape masked = {}'.format(nearest_neighbor_indices.shape,nni.shape))

        convolved_density = np.zeros(n_crs,dtype = density_cart.dtype)
        convolved_density[mask] = convolved_density_cart[nni[:,0],nni[:,1],nni[:,2]].real
        #convolved_density[:] = convolved_density_cart[:].real
        return convolved_density

    def convolve_density_with_gaussian2(self,crs_density,sigma):
        '''
        Convolves a density in self._densities with a gaussian.
        '''
        n_crs = self.header_part.n_crs        
        n_steps = 2**(np.log2(2*n_crs.max()) + 0.5).astype(int)
        grid_bounds = np.max(np.abs(self.bounds.xyz),axis=-1)
        step_size = grid_bounds/(n_steps//2)
        
        steps = [np.arange(-n_steps//2+1,n_steps//2+1) * step_size[i] for i in range(3)]        
        cart_grid = GridFactory.construct_grid('uniform',steps).array
        density_cart = self.sample_density_xyz(crs_density,cart_grid)
        
        convolved_density_cart = convolve_with_gaussian_cart(density_cart,cart_grid,sigma)
        log.info('convolved .shape = {}'.format(convolved_density_cart.shape))

        crs_grid = GridFactory.construct_grid('uniform',[np.arange(n) for n in n_crs]).array
        xyz_grid = self.crs2xyzCoord(crs_grid.reshape(-1,3)).reshape(crs_grid.shape)
        mask = np.prod((xyz_grid >= -grid_bounds[None,None,None,:]) & (xyz_grid <= grid_bounds[None,None,None,:]),axis = -1).astype(bool)
       # log.info('mas is true = {}'.format(mask.all()))
        #log.info("{} masked ccordinates".format(sum(~mask.flatten())))
        nearest_neighbor_indices = ((xyz_grid+grid_bounds[None,None,None,:])//step_size[None,None,None,:]).astype(int)
        #log.info('{} indices bigger than 128!'.format(np.sum(nearest_neighbor_indices>=128)))
        #log.info('mask shape = {}'.format(mask.shape))
        nni = nearest_neighbor_indices[mask]
        #log.info('{} indices bigger than 128!'.format(np.sum(nni>=128)))
        #log.info('nni shape unmasked = {}, nni shape masked = {}'.format(nearest_neighbor_indices.shape,nni.shape))

        convolved_density = np.zeros(n_crs,dtype = density_cart.dtype)
        convolved_density[mask] = convolved_density_cart[nni[:,0],nni[:,1],nni[:,2]].real
        #convolved_density[:] = convolved_density_cart[:].real
        return convolved_density
    
    def extract_density(self,name,grid,atom_neighbour_shells = 2, gaussian_sigma = False ):
        '''
        Extracts the given density profile ("fo","fc","two_fofc","fofc") 
        '''
        integrator = SphericalIntegrator(grid)
        crs_density = self._densities[name]
        density_mask = self.generate_density_mask(atom_neighbour_shells = atom_neighbour_shells)
        log.info('densitymask dtype = {}'.format(density_mask.dtype))
        crs_density[~density_mask] = 0
        if not isinstance(gaussian_sigma ,bool):
            crs_density = self.convolve_density_with_gaussian(crs_density,float(gaussian_sigma))


        rs = grid[:,0,0,0]
        r_mask = rs<self.max_r
        cart_grid = spherical_to_cartesian(grid)
        
        
        density = np.zeros(cart_grid.shape[:-1])
        cart_support = cart_grid[r_mask,...]
        support_shape = cart_support.shape
        crs_support = self.xyz2crsCoord(cart_support.reshape(-1,3)).reshape(*support_shape)
        crs_support,mask = self.return_valid_crs_coords(crs_support,return_mask=True)
        supported_density_part = crs_density[crs_support[:,0],crs_support[:,1],crs_support[:,2]]

        temp = density[r_mask]
        temp[mask] = supported_density_part
        density[r_mask] = temp
        #density *= np.sqrt(1/integrator.integrate(density**2))
        return density

class DensityExtractor_old:
    def __init__(self,pdb_id):
        cwd = os.getcwd()
        with tempfile.TemporaryDirectory() as tmpdirname:
            os.chdir(tmpdirname)
            log.info('Start loading files from PDB. This might take a minute..')
            densityAnalysisObj = pdb_eda.densityAnalysis.fromPDBid(pdb_id,downloadFile = False,pdbi=True,mmcif=False,pdbbio=True)
            assert densityAnalysisObj != 0, 'given pdb_id = {} dies not exist or does not contain 2fo-fc density data'.format(pdb_id)
            self._densities= DictNamespace(
                fo = densityAnalysisObj.fo.density.T,
                fc = densityAnalysisObj.fc.density.T,
                two_fcfo = densityAnalysisObj.densityObj.density.T,
                fcfo = densityAnalysisObj.diffDensityObj.density.T
            )
            h=densityAnalysisObj.densityObj.header
            self.header_part = DictNamespace(
                n_crs = np.asarray(h.ncrs),
                grid_length = np.asarray(h.gridLength),
                origin = np.asarray(h.origin),
                xyz_interval = np.asarray(h.xyzInterval),
                crs_start = np.asarray(h.crsStart),
                map_to_xyz = np.asarray(h.map2xyz),
                map_to_crs = np.asarray(h.map2crs),
                ortho_mat = np.asarray(h.orthoMat),
                inv_ortho_mat = np.asarray(h.deOrthoMat),
                alpha = h.alpha,
                beta = h.beta,
                gamma = h.gamma        
            )
            self._atom_coords_xyz = np.asarray(densityAnalysisObj.asymmetryAtomCoords)

        self.atom_coords = DictNamespace(
            xyz = self._atom_coords_xyz,
            crs = self.xyz2crsCoord(self._atom_coords_xyz)
        )
        self.bounds = DictNamespace(
            xyz = np.array(tuple((self.atom_coords.xyz[...,n].min(),self.atom_coords.xyz[...,n].max()) for n in range(3))),
            crs = np.array(tuple((self.atom_coords.crs[...,n].min(),self.atom_coords.crs[...,n].max()) for n in range(3)))            
        )
        self.max_r = np.linalg.norm(self.atom_coords.xyz,axis= -1).max()
        os.chdir(cwd)
                
    def crs2xyzCoord(self,crs):
        '''
        Convert the crs coordinates into xyz coordinates.
        Assumes crs is of shape (N,3) where N is the number of coordinate sets.
        '''
        h = self.header_part
        crs = np.atleast_2d(crs)
        if h.alpha == h.beta == h.gamma == 90:
            xyz = crs[:,h.map_to_xyz]*h.grid_length[None,:]+h.origin[None,:]
            
        else:
            xyz = np.dot(h.ortho_mat, ((crs[:,h.map_to_xyz]+h.crs_start[None,h.map_to_xyz])/h.xyz_interval[None,:]).T).T
        return np.squeeze(xyz)
    
    def xyz2crsCoord(self,xyz):
        '''
        Converts xyz to crs coordinates. 
        Assumes xyz is of shape (N,3) where N is the number of coordinate sets.
        '''
        h = self.header_part
        xyz = np.atleast_2d(xyz)
        if h.alpha == h.beta == h.gamma == 90:
            crsGridPos =  np.rint((xyz - h.origin[None,:])/h.grid_length[None,:])[:,h.map_to_crs]
        else:
            fraction = np.dot(h.inv_ortho_mat, xyz.T).T
            crsGridPos = (np.rint(fraction*h.xyz_interval[None,:]) - h.crs_start[None,h.map_to_xyz])[:,h.map_to_crs]
        return np.squeeze(crsGridPos).astype(int)

    def sample_density_xyz(self,crs_density,points,fill_value = 0.0):
        '''
        Samples density at given xyz coordinates.
        '''
        shape = points.shape
        crs_points = self.xyz2crsCoord(points.reshape(-1,3))
        sampled_density = self.sample_density_crs(crs_density,crs_points,fill_value = fill_value).reshape(shape[:-1])
        return sampled_density
    
    def sample_density_crs(self,crs_density,points,fill_value = 0.0):
        '''
        Samples density at given crs coordinates.
        fills defaut value for crs coordinates which are smaller than 0 and bigger than self.n_crs.
        '''
        points = np.atleast_2d(points)
        shape = points.shape
        log.info('shape crs = {}'.format(shape))
        valid_crs_points , mask = self.return_valid_crs_coords(points.reshape(-1,3),return_mask= True)

        sampled_density=np.full(points.shape[:-1],fill_value,dtype = crs_density.dtype)
        log.info('sample density shape = {}'.format(sampled_density.shape))
        sampled_density[mask] = crs_density[valid_crs_points[:,0],valid_crs_points[:,1],valid_crs_points[:,2]]
        return sampled_density.reshape(shape[:-1])
    def return_valid_crs_coords(self,coords,return_mask = False):
        '''
        Returns all coordinates tripples that satisfy  0<= val[i] < heade.n_crs[i] for i in 0,1,2.
        '''
        valid_mask = (np.prod( (coords < self.header_part.n_crs[None,:]) * (coords >= 0),axis=-1)).astype(bool)
        coords = coords[valid_mask,:]
        if return_mask:
            return coords,valid_mask
        else:
            return coords
        
    def generate_density_mask(self,atom_neighbour_shells=2):
        '''
        Generates a mask around each atom of the PDB structure. (To get rid of symmetry parts of the density fuction.)
        :param atom_neighbor_shells: Number of cubical shells around a given crs coordinate which to include in the density
        :type int: 
        '''
        crs = self.atom_coords.crs
        neighbours_1d = np.arange(-atom_neighbour_shells ,atom_neighbour_shells + 1)
        neighbours_3d = GridFactory.construct_grid('uniform',[neighbours_1d,neighbours_1d,neighbours_1d]).flatten().array
        valid_grid_positions = np.unique(self.return_valid_crs_coords((crs[:,None,:] + neighbours_3d[None,...]).reshape(-1,3)),axis = 0)
        mask = np.full(self.header_part.n_crs,False)
        mask[valid_grid_positions.T[0],valid_grid_positions.T[1],valid_grid_positions.T[2]]=True
        return mask

    def convolve_density_with_gaussian(self,crs_density,sigma):
        '''
        Convolves a density in self._densities with a gaussian.
        '''
        n_crs = self.header_part.n_crs        
        n_steps = 2**(np.log2(2*n_crs.max()) + 0.5).astype(int)
        grid_bounds = np.max(np.abs(self.bounds.xyz),axis=-1)
        step_size = grid_bounds/(n_steps//2)
        
        steps = [np.arange(-n_steps//2+1,n_steps//2+1) * step_size[i] for i in range(3)]        
        cart_grid = GridFactory.construct_grid('uniform',steps).array
        density_cart = self.sample_density_xyz(crs_density,cart_grid)
        
        convolved_density_cart = convolve_with_gaussian_cart(density_cart,cart_grid,sigma)
        log.info('convolved .shape = {}'.format(convolved_density_cart.shape))

        crs_grid = GridFactory.construct_grid('uniform',[np.arange(n) for n in n_crs]).array
        xyz_grid = self.crs2xyzCoord(crs_grid.reshape(-1,3)).reshape(crs_grid.shape)
        mask = np.prod((xyz_grid >= -grid_bounds[None,None,None,:]) & (xyz_grid <= grid_bounds[None,None,None,:]),axis = -1).astype(bool)
       # log.info('mas is true = {}'.format(mask.all()))
        #log.info("{} masked ccordinates".format(sum(~mask.flatten())))
        nearest_neighbor_indices = ((xyz_grid+grid_bounds[None,None,None,:])//step_size[None,None,None,:]).astype(int)
        #log.info('{} indices bigger than 128!'.format(np.sum(nearest_neighbor_indices>=128)))
        #log.info('mask shape = {}'.format(mask.shape))
        nni = nearest_neighbor_indices[mask]
        #log.info('{} indices bigger than 128!'.format(np.sum(nni>=128)))
        #log.info('nni shape unmasked = {}, nni shape masked = {}'.format(nearest_neighbor_indices.shape,nni.shape))

        convolved_density = np.zeros(n_crs,dtype = density_cart.dtype)
        convolved_density[mask] = convolved_density_cart[nni[:,0],nni[:,1],nni[:,2]].real
        #convolved_density[:] = convolved_density_cart[:].real
        return convolved_density

    def convolve_density_with_gaussian2(self,crs_density,sigma):
        '''
        Convolves a density in self._densities with a gaussian.
        '''
        n_crs = self.header_part.n_crs        
        n_steps = 2**(np.log2(2*n_crs.max()) + 0.5).astype(int)
        grid_bounds = np.max(np.abs(self.bounds.xyz),axis=-1)
        step_size = grid_bounds/(n_steps//2)
        
        steps = [np.arange(-n_steps//2+1,n_steps//2+1) * step_size[i] for i in range(3)]        
        cart_grid = GridFactory.construct_grid('uniform',steps).array
        density_cart = self.sample_density_xyz(crs_density,cart_grid)
        
        convolved_density_cart = convolve_with_gaussian_cart(density_cart,cart_grid,sigma)
        log.info('convolved .shape = {}'.format(convolved_density_cart.shape))

        crs_grid = GridFactory.construct_grid('uniform',[np.arange(n) for n in n_crs]).array
        xyz_grid = self.crs2xyzCoord(crs_grid.reshape(-1,3)).reshape(crs_grid.shape)
        mask = np.prod((xyz_grid >= -grid_bounds[None,None,None,:]) & (xyz_grid <= grid_bounds[None,None,None,:]),axis = -1).astype(bool)
       # log.info('mas is true = {}'.format(mask.all()))
        #log.info("{} masked ccordinates".format(sum(~mask.flatten())))
        nearest_neighbor_indices = ((xyz_grid+grid_bounds[None,None,None,:])//step_size[None,None,None,:]).astype(int)
        #log.info('{} indices bigger than 128!'.format(np.sum(nearest_neighbor_indices>=128)))
        #log.info('mask shape = {}'.format(mask.shape))
        nni = nearest_neighbor_indices[mask]
        #log.info('{} indices bigger than 128!'.format(np.sum(nni>=128)))
        #log.info('nni shape unmasked = {}, nni shape masked = {}'.format(nearest_neighbor_indices.shape,nni.shape))

        convolved_density = np.zeros(n_crs,dtype = density_cart.dtype)
        convolved_density[mask] = convolved_density_cart[nni[:,0],nni[:,1],nni[:,2]].real
        #convolved_density[:] = convolved_density_cart[:].real
        return convolved_density
    
    def extract_density(self,name,grid,atom_neighbour_shells = 2, gaussian_sigma = False ):
        '''
        Extracts the given density profile ("fo","fc","two_fofc","fofc") 
        '''
        integrator = SphericalIntegrator(grid)
        crs_density = self._densities[name]
        density_mask = self.generate_density_mask(atom_neighbour_shells = atom_neighbour_shells)
        log.info('densitymask dtype = {}'.format(density_mask.dtype))
        crs_density[~density_mask] = 0
        if not isinstance(gaussian_sigma ,bool):
            crs_density = self.convolve_density_with_gaussian(crs_density,float(gaussian_sigma))


        rs = grid[:,0,0,0]
        r_mask = rs<self.max_r
        cart_grid = spherical_to_cartesian(grid)
        
        
        density = np.zeros(cart_grid.shape[:-1])
        cart_support = cart_grid[r_mask,...]
        support_shape = cart_support.shape
        crs_support = self.xyz2crsCoord(cart_support.reshape(-1,3)).reshape(*support_shape)
        crs_support,mask = self.return_valid_crs_coords(crs_support,return_mask=True)
        supported_density_part = crs_density[crs_support[:,0],crs_support[:,1],crs_support[:,2]]

        temp = density[r_mask]
        temp[mask] = supported_density_part
        density[r_mask] = temp
        #density *= np.sqrt(1/integrator.integrate(density**2))
        return density
        
        



import numpy as np
import os
os.chdir('/gpfs/exfel/theory_group/user/berberic/Programs/xframe')
from xframe.database.euxfel import default_DB
from xframe.presenters.matplolibPresenter import plot1D

db = default_DB()

base_path = '/gpfs/exfel/theory_group/user/berberic/MTIP/reconstructions/20_8_2022/run_0/'
path = base_path + 'reconstruction_data.h5'
data = db.load(path)
r = data['reconstruction_results']
errors = np.moveaxis(np.array([d['error_dict']['deg2_invariant_l2_diff'] for d in r.values()]),2,0)

order =6
fig = plot1D.get_fig(errors[order,:,2:],y_scale = 'log')
fig_path = base_path + 'order_{}_Bl_errors.matplotlib'.format(order)
db.save(fig_path,fig)



import scipy.constants as constants

h=constants.physical_constants['Planck constant in eV s'][0]
hbar=constants.physical_constants['Planck constant over 2 pi in eV s'][0]
c=constants.c
standardLength=1e-10



import abc
import logging
log=logging.getLogger('root')
from xframe.interfaces import DependencyMeta

class HDF5Interface(abc.ABC):
    _external_dependency_name_ = 'hdf5'
    @abc.abstractmethod
    def save(path,data: dict):
        log.error('HDF5 dependency not injected. Is h5py python package installed ?')
    @abc.abstractmethod
    def load(path):
        log.error('HDF5 dependency not injected. Is h5py python package installed ?')

class HDF5Dependency(HDF5Interface,metaclass=DependencyMeta):
    pass
class YAMLInterface(abc.ABC):
    _external_dependency_name_ = 'yaml'
    @abc.abstractmethod
    def save(path,data: dict):
        log.error('YAML dependency not injected. Is the ruamel python package installed ? ')
    @abc.abstractmethod
    def load(path):
        log.error('YAML dependency not injected. Is the ruamel python package installed ?')
    @abc.abstractmethod
    def format_settings_dict(dict):
        log.error('YAML dependency not injected. Is the ruamel python package installed ?')

class YAMLDependency(YAMLInterface,metaclass = DependencyMeta):
    pass
class VTKInterface(abc.ABC):
    _external_dependency_name_ = 'vtk'
    @abc.abstractmethod
    def save(datasets, grid, file_path,dset_names = 'data', grid_type='cartesian'):
        log.error('VTK dependency not injected. Is the vtk package installed ')
    @abc.abstractmethod
    def load(path):
        log.error('VTK dependency not injected. Is the vtk python package installed ?')

class VTKDependency(VTKInterface,metaclass = DependencyMeta):
    pass
class PDBInterface(abc.ABC):
    _external_dependency_name_ = 'pdb'
    @abc.abstractmethod
    def load(pdb_id : str):
        log.error('PDB dependency not injected. Is the pdb_eda python package installed ? ')
    @abc.abstractmethod
    def save(path,data):
        log.error('PDB dependency not injected. Is the pdb_eda python package installed ?')

class PDBDependency(PDBInterface,metaclass = DependencyMeta):
    pass
class OpenCVInterface(abc.ABC):
    _external_dependency_name_ = 'opencv'
    @abc.abstractmethod
    def load(path):
        log.error('OpenCV dependency not injected. Is the opencv-python python package installed ? ')
    @abc.abstractmethod
    def save(path,data):
        log.error('OpenCV dependency not injected. Is the opencv-python python package installed ?')

class OpenCVDependency(OpenCVInterface, metaclass = DependencyMeta):
    pass



import inspect
import os
import ctypes
from ctypes.util import find_library
import time
import numpy as np
import numpy.ma as mp
from scipy.interpolate import griddata
from functools import partial
from itertools import repeat
from types import SimpleNamespace
import logging
from io import StringIO
import sys
from contextlib import contextmanager

sa = 'Shared_Array module'

log=logging.getLogger('root')

doublePrecision=1e-16

def xprint(txt):
    if log.level > logging.INFO:
        print(txt)
    log.info(txt)
    
    
def hash_numpy_in_tuple(_tuple):       
    tmp=[]
    for item in _tuple:
        if isinstance(item,np.ndarray):
            tmp.append(hash(item.data.tobytes()))
        else:
            tmp.append(item)
    return tuple(tmp)
    
    
def load_clib(libraryName, mode = None):
    libraryPath = find_library(libraryName)
    if libraryPath is None:
        log.info('Did not find library with name "{}". Try interpreting it as the librarys path.'.format(libraryName))
        libraryPath=libraryName
    try:
        if mode is None:
            library = ctypes.CDLL(libraryPath)
        else:
            library = ctypes.CDLL(libraryPath, mode)
    except OSError as e:
        log.error(e)
        log.error('current basepath is: {}'.format(os.getcwd()))
        raise e
    return library

def plugVariableIntoFunction(function,variableSpecifier):
    variable=variableSpecifier[0]
    varPosition=variableSpecifier[1]
    def newFunction(*args,**kwargs):
        arguments=args[0:varPosition]+(variable,)+args[varPosition:]
        output=function(*arguments,**kwargs)
        return output
    
    remainingParameters=tuple(inspect.signature(function).parameters.values())[:varPosition]+tuple(inspect.signature(function).parameters.values())[varPosition+1:]
    newFunction.__signature__=inspect.signature(newFunction).replace(parameters=remainingParameters)
    return newFunction

def appendArgumentsToFunction(function,*args,**kwargs):
    def newFunction(firstArgument):
        output=function(firstArgument,*args,**kwargs)
        return output
    #applies the parameter signature of function to newFunction
    firstParameter=tuple(inspect.signature(function).parameters.values())[0]
    newFunction.__signature__=inspect.signature(newFunction).replace(parameters=[firstParameter])
    return newFunction

def partialSums(array1D):
    def sum(elementNumber):
        return np.sum(array1D[:elementNumber+1])
    elementNumbers=np.arange(len(array1D))
    listOfSums=np.array(list(map(sum,elementNumbers)))
    return listOfSums

def listToDict(list):
    dict={list[i]:list[i+1] for i in range(0,len(list),2)}
    return dict

def listOfTuplesToDict(list):
    dict={tuple[0]:tuple[1] for tuple in list}
    return dict

def invertNumpyIndex(index):
    invertedIndex=np.zeros(len(index),dtype=int)
    for pos,i in enumerate(index):
        invertedIndex[i]=pos
    return invertedIndex

def getFunctions(_object, _type='list',recoursive=False):
    functions=[]
    if recoursive:
        submodules=inspect.getmembers(_object,predicate=inspect.ismodule)
        for module in submodules:
            functions+=inspect.getmembers(module[1],predicate=inspect.isfunction)
    else:
        functions+=inspect.getmembers(_object,predicate=inspect.isfunction)
    if _type=='dict':
        functions=listOfTuplesToDict(functions)
    return functions

def getFunctionNames(_object):
    functions=getFunctions(_object)
    functionNames=[tuple[0] for tuple in functions]
    return functionNames

def getMethods(_object, _type='list'):
    methods=inspect.getmembers(_object,predicate=inspect.ismethod)
    if _type=='dict':
        methods=listOfTuplesToDict(methods)
    return methods

def getMethodNames(_object):
    methods=getMethods(_object)
    methodNames=[tuple[0] for tuple in methods]
    return methodNames

def measureTime(func):
    def wrapper(*args,**kwargs):
#        log.info('kwargs={}'.format(kwargs))
        startTime=time.time()
        output=func(*args,**kwargs)
        duration=time.time()-startTime
        xprint('function call took {} seconds'.format(duration))
        return output
    return wrapper

def isIterable(object):
    try:
        iter(object)
    except Exception:
        return False
    else:
        return True
def generateMesh(*args,joiningDimension=2):
    args=list(map(np.asarray,args))
    mesh=np.stack(np.meshgrid(*args,indexing='ij'),joiningDimension)
    return mesh

def meshstack(array1,array2):
    results=[]
    for a,b in zip(array1,array2):
        temp=np.full((len(b),len(a)),a)
        result=np.concatenate((temp,np.expand_dims(b,axis=1)),axis=1)
        results.append(result)
    return results

def flattenList(_list):
    elements=[]
    for sub_list in _list:
        if isinstance(sub_list,list): 
            newElements=flattenList(sub_list)
            elements +=  newElements
        else:
            elements += [sub_list]
    return elements



def uniformGridGetStepSizes(gridArray):
    gridShape=gridArray.shape
#    log.info('grid shape={}'.format(gridShape))
    if gridArray.dtype==np.object:
        gridDimension=len(gridShape)
    else:
        gridDimension=gridShape[-1]

#    log.info('grid Dimension={}'.format(gridDimension))

    secondPointIndices=np.eye(gridDimension,dtype=np.int)
    thirdPointIndices=secondPointIndices*2
    stepSizes=gridArray[tuple(thirdPointIndices)]-gridArray[tuple(secondPointIndices)]
    indices=np.repeat(np.expand_dims(np.arange(gridDimension),0),2,axis=0)
#    log.info('indices={}'.format(indices))
#    log.info('steps={}'.format(np.array(tuple(stepSizes))))
    stepSizes=np.array(tuple(stepSizes))[tuple(indices)]
#    log.info('stepsizes dtype={}'.format(stepSizes.dtype))
    return stepSizes

def uniformGridToFunction(dataPoints,gridSpec):
    #axisTypes is an np array of length grid Dimension.
    origin=gridSpec[0]
    stepSizes=gridSpec[1]
    log.info('origin={} stepSizes={}'.format(origin,stepSizes))
    gridShape=dataPoints.shape
    gridShapeArray=np.array(gridShape,dtype=np.int)
    gridDimension=len(gridShape)

#    if isinstance(axisTypes,bool):
#        axisTypes=np.zeros(gridDimension)
#    else:
#        axisTypes=np.asarray(axisTypes)

#    axisIsPeriodic= (axisTypes==1)
#    axisHasNoEndpoint= (axisTypes==2)

    log.info('gridDimension={} step Sizes={}'.format(gridDimension,stepSizes))
    def dataFunction(*args):        
        point=np.array([args[index] for index in range(gridDimension)])
        point=point-origin
#        log.info('point={}'.format(point))
        dataIndex=np.floor(point/stepSizes).astype(np.int)

#        dataIndex[axisIsPeriodic]=dataIndex[axisIsPeriodic]%gridShapeArray[axisIsPeriodic]
 #       noEndpointIndex=dataIndex[axisHasNoEndpoint]
 #       dataIndex[axisHasNoEndpoint]=np.where(noEndpointIndex==gridShape,noEndpointIndex-1,noEndpointIndex)
        
        isOutsideOfGrid= (gridShape-dataIndex<=0).any()
        if isOutsideOfGrid:
            value=0
            log.warning('grid Function is called at point {}*{} outside of grid. Returning 0!'.format(dataIndex,stepSizes))
        else:
            value=dataPoints[tuple(dataIndex)]        
        return value
    return dataFunction
    
def getGridByXYValues(xValues,yValues):
    def xfunc(*args):
        return xValues
    def yfunc(*args):
        return yValues
    gfact=gridFactory()
    newGrid=gfact.constructGrid([xfunc,yfunc],'uniform')
    return newGrid

def restoreOldParameters(oldFunction,newFunction):
    oldParameters=inspect.signature(oldFunction).parameters.values()
    newFunction.__signature__=inspect.signature(newFunction).replace(parameters=oldParameters)
    return newFunction

def selectElementOfFunctionOutput_decorator(function,elementIndex):
    def newFunction(*args,**kwargs):
        output=function(*args,**kwargs)
        log.info('len output ={}'.format(len(output)))
        log.info('shapes of outputs ={}'.format([x.shape for x in output]))
        return output[elementIndex]

    #resore original parameters
    newFunction=restoreOldParameters(function,newFunction)
    return newFunction

def getFirstElement(subscriptable):
    return subscriptable[0]

def getLastElement(subscriptable):
    return subscriptable[-1]

def swapElements(array,index1,index2):
    array=np.asarray(array)
    value1=array[index1]
    array[index1]=array[index2]
    array[index2]=value1
    return array

def concatenateDecorators(decoratorList):
    def newDecorator(function):
        newFunction=function
        for decorator in decoratorList:
            newFunction=decorator(newFunction)
        return newFunction
    return newDecorator
    
def getArrayOfArray(array):
#    log.info('input array = {}'.format(array))
    shape=array.shape
    if len(shape)!=1:
        array=array.reshape(-1,shape[-1])
        
    arrayOfArrays=np.frompyfunc(array.__getitem__,1,1)(range(array.shape[0]))
    arrayOfArrays=arrayOfArrays.reshape(shape[:-1])
    return arrayOfArrays

def ArrayOfArrays2Array(array):
    shape=array.shape
    new_array=np.array(tuple(array.flatten()))
    return newArray

def optional_arg_decorator(fn):
    def wrapped_decorator(*args,**kwargs):
        if len(kwargs) == 0 and callable(args[0]):
            return fn(args[0])
        else:
            def real_decorator(decoratee):
                return fn(decoratee, **kwargs)
            return real_decorator
    return wrapped_decorator


#########
##MP utils##


def splitMPArguments(argArrays,nProcesses,mode='sequential',split_together=False):
    #log.info('argArrays={}'.format(argArrays))
    if isinstance(argArrays,np.ndarray):
    #    log.info('yep')
        argArrays=[argArrays]
    indiceArrays=[np.arange(len(array),dtype=int) for array in argArrays]
    arg_legths_equal= len(set(tuple(len(array) for array in argArrays))) == 1
    dtype_list = [arg.dtype for arg in argArrays]
    n_args = len(argArrays)
    #log.info('len argArrays ={}'.format(len(argArrays)))
    if split_together:
        try:
            assert arg_legths_equal,'split_together requires all argArrays to have the same length but the length are {}'.format(tuple(len(array) for array in argArrays))           
        except AssertionError as e:
            log.error(e)
            raise
        arguments=tuple(argArrays)
        argIndices=indiceArrays
       
    else:
        #    log.info('indiceArrays={}'.format(indiceArrays))
        argIndices=tuple(reversed(tuple(dataset.flatten().astype(np.int) for dataset in np.meshgrid(*reversed(indiceArrays)) )))
        arguments=tuple(argument[indices] for argument,indices in zip(argArrays,argIndices))
    nArguments=len(arguments[0])
    #print('arguments = {}'.format(argArrays))    
    log.info('nArgs = {} nProcess {}'.format(nArguments,nProcesses))
    if nArguments<nProcesses:
        nProcesses=nArguments
    argPerProcess=nArguments//nProcesses    

    #    log.info('nArguments={}, nProcesses={}'.format(nArguments,nProcesses))
    #print(mode)

    if mode=='sequential':    
        splitPoints=[int(argPerProcess*process) for process in range(1,nProcesses)]
        splittedArguments=tuple(np.split(argument,splitPoints) for argument in arguments)
        splittedIndices=tuple(np.split(argIndice,splitPoints) for argIndice in argIndices)
    if mode=='modulus':
        def splitArgument(argument):
            dtype=argument.dtype
            def returnSlizedArgument(processID):
                arg_slice = np.array(argument[processID::nProcesses],dtype=dtype)
 #               log.info('slice shape = {}'.format(arg_slice.shape))
                return arg_slice
            splittedArgument=np.array(tuple(map(returnSlizedArgument,range(nProcesses))),dtype=np.object)
            #log.info('splittedArgument={}'.format(splittedArgument))
            return splittedArgument
        
        splittedArguments=tuple(splitArgument(argument) for argument in arguments)
        splittedIndices=tuple(splitArgument(argIndice) for argIndice in argIndices)
    #      log.info('dtype={} splitted Indices=\n {}'.format(splittedIndices.dtype,splittedIndices))
    
    arguments_per_process = list(zip(*splittedArguments))
    indices_per_process = list(zip(*splittedIndices))    
    return {'arguments':arguments_per_process,'indices':indices_per_process}


def splitMPArguments_old(argArrays,nProcesses,mode='sequential'):    
    if isinstance(argArrays,np.ndarray):
    #    log.info('yep')
        argArrays=[argArrays]
        
   
    indiceArrays=[np.arange(len(array),dtype=int) for array in argArrays]
#    log.info('indiceArrays={}'.format(indiceArrays))

    if len(argArrays) == 1:

        arguments=(argArrays[0],)
        argIndices=(np.arange(len(arguments)),)
    else:
        arguments=tuple(reversed(tuple(dataset.flatten() for dataset in np.meshgrid(*reversed(argArrays)) )))
        argIndices=tuple(reversed(tuple(dataset.flatten().astype(np.int) for dataset in np.meshgrid(*reversed(indiceArrays)) )))
    nArguments=arguments[0].shape[0]
    if nArguments<nProcesses:
        nProcesses=nArguments
    argPerProcess=np.floor(nArguments/nProcesses)

#    log.info('nArguments={}, nProcesses={}'.format(nArguments,nProcesses))

    if mode=='sequential':    
        splitPoints=[int(argPerProcess*process) for process in range(1,nProcesses) ]
        splittedArguments=np.array([np.split(argument,splitPoints) for argument in arguments],dtype=np.object)
        splittedIndices=np.array([np.split(argIndice,splitPoints) for argIndice in argIndices],dtype=np.object)
    if mode=='modulus':
        def splitArgument(argument):
            dtype=argument.dtype
            def returnSlizedArgument(processID):
                arg_slice = np.array(argument[processID::nProcesses],dtype=dtype)
 #               log.info('slice shape = {}'.format(arg_slice.shape))
                return arg_slice
            splittedArgument=np.array(tuple(map(returnSlizedArgument,range(nProcesses))),dtype=np.object)
            return splittedArgument
        
        splittedArguments=np.array([splitArgument(argument) for argument in arguments])
        splittedIndices=np.array([splitArgument(argIndice) for argIndice in argIndices])
  #      log.info('dtype={} splitted Indices=\n {}'.format(splittedIndices.dtype,splittedIndices))
    argumentsPerProcessList=np.swapaxes(splittedArguments,0,1)
    indicesPerProcessList=np.swapaxes(splittedIndices,0,1)
    #    indicesPerProcessList2=[np.array(indices,dtype=np.int) for indices in indicesPerProcessList]
    #log.info('arguments per process=\n{}'.format(argumentsPerProcessList))
    return {'arguments':argumentsPerProcessList,'indices':indicesPerProcessList}

def assembleResults(resultDict,mode='sequential'):
#    log.info('result 0 shape={}'.format(results[0].shape))
    nProcesses=len(resultDict)
    nCallsPerProcess=np.array(tuple(len(resultDict[processID]) for processID in range(nProcesses)))
    nTotalCalls=np.sum(nCallsPerProcess)
#    log.info('nProcesses={}'.format(nProcesses))
#    log.info('nCallsPerProcess={}'.format(nCallsPerProcess))
#    log.info('nResults={}'.format(nResults))
    
    if mode=='sequential':        
        result=()
        for processID in range(nProcesses):
  #            log.info('result part={}'.format(resultDict[processID]))
            result+=tuple(resultDict[processID])
        
    elif mode=='modulus':
        result=list(range(nTotalCalls))
        for processID in range(nProcesses):
            result[processID::nProcesses]=list(resultDict[processID])
        
    result=np.array(result)
    
    return result


def get_mp_function():
    pass


def getMPFunction(function,arguments,args=[],callWithMultipleArguments=True,output_file = False,**kwargs):
    repeatedArguments=[repeat(arg) for arg in args]
    #log.info('arguments={}'.format(arguments))
    if len(arguments)>0:
        nParts=(len(arguments[0]),)
    else:
        nParts=(1,)
        arguments=[[None]]
    #log.info('nParts={}'.format(nParts))
#    log.info('newShape={}'.format(newShape))
    #log.info('repeatedArguments={}'.format(repeatedArguments))
    #log.info('call with multiple args={}'.format(callWithMultipleArguments))
    if callWithMultipleArguments:        
        def mpFunction(queue,processID):
            kwargs["processID"] = processID
            result=function(*arguments,*args,**kwargs)
            #            log.info('result type={}'.format(type(result)))
            if isinstance(result,(list,tuple)):
                result=np.asarray(result)

            result_dict={processID:result}
            #           log.debug('processID type={}'.format(type(processID)))
            queue.put(result_dict,False)
    else:
        def mpFunction(queue,processID):
            kwargs["processID"] = processID
            #log.info(next(zip(*arguments)))
            result=list(function(*argument,*args,**kwargs) for argument in zip(*arguments))
            if len(result)>0:
                all_results_are_numpy_arrays = np.array([isinstance(r,np.ndarray) for r in result]).all()
                if all_results_are_numpy_arrays:
                    all_results_have_same_shape = np.array([r.shape==result[0].shape for r in result]).all()
                    if all_results_have_same_shape:
                        result = np.concatenate(result)
            #results={processID:result_array.reshape(nParts+result_array.shape[1:])}
            results={processID:result}
            #        log.info('results=\n{}'.format(results))
            #        log.info('putting result in queue')
            queue.put(results,False)
            #        log.info('{} elements are in queue'.format(queue.qsize())
    return mpFunction

def get_mp_function_SharedArray(function,arguments,output_shape,args=[],callWithMultipleArguments=True,output_file = False,output_indices=slice(None),**kwargs):
    repeatedArguments=[repeat(arg) for arg in args]
    #log.info('arguments={}'.format(arguments))
    if len(arguments)>0:
        nParts=(len(arguments[0]),)
    else:
        nParts=(1,)
        arguments=['mock_argument']
    #log.info('nParts={}'.format(nParts))
#    log.info('newShape={}'.format(newShape))
    #log.info('repeatedArguments={}'.format(repeatedArguments))
    #log.info('call with multiple args={}'.format(callWithMultipleArguments))
    if callWithMultipleArguments:        
        def mpFunction(processID):
            output_file = sa.attach('shm://shared_output').reshape(output_shape)            
            output_file[output_indices]=function(*arguments,*args,**kwargs)
            del(output_file)
    else:
        def mpFunction(processID):
            output_file = sa.attach('shm://shared_output').reshape(output_shape)
            output_file[output_indices]=np.array(list(function(argument,*args,**kwargs) for argument in arguments))
            del(output_file)

                #        log.info('{} elements are in queue'.format(queue.qsize())
    return mpFunction

def get_mp_function_SharedArray_multi(function,arguments,output_names,output_shape,args=[],callWithMultipleArguments=True,output_file = False,output_indices=slice(None),**kwargs):
    repeatedArguments=[repeat(arg) for arg in args]
    #log.info('arguments={}'.format(arguments))
    if len(arguments)>0:
        nParts=(len(arguments[0]),)
    else:
        nParts=(1,)
        arguments=['mock_argument']
    #log.info('nParts={}'.format(nParts))
    #    log.info('newShape={}'.format(newShape))
    #log.info('repeatedArguments={}'.format(repeatedArguments))
    #log.info('call with multiple args={}'.format(callWithMultipleArguments))
    
    if callWithMultipleArguments:        
        def mpFunction(processID):            
            output_files = [sa.attach('shm://'+ name).reshape(output_shape) for name in output_names]
            outputs = function(*arguments,*args,**kwargs)
            #log.info('len outputs = {}'.format(len(outputs)))
            #log.info('len outputs = {}'.format(len(output_files)))
            #log.info('len outputs = {}'.format(len(output_indices)))
            out_ids = output_indices
            for o_file,out in zip(output_files,outputs):
                #log.info(' shared array file dtype = {}'.format(o_file.dtype))
                #log.info(' out dtype = {}'.format(out.dtype))
                #log.info(' out shape = {}'.format(out.shape))
                ##if out.dtype == np.dtype(bool):
                #    log.info('shared array file dtype = {}'.format(o_file.dtype))
                #    log.info('{} % = {}'.format(np.sum(out)/np.prod(out.shape)))
                o_file[out_ids]=out
                #log.info('dtypes = {}'.format([o_file.dtype,out.dtype]))
                del(o_file)
    else:
        def mpFunction(processID):
            output_files = [sa.attach('shm://'+ name).reshape(output_shape) for name in output_names]
            outputs = list(zip(*list(function(argument,*args,**kwargs) for argument in arguments)))
            for o_file,out,out_ids in zip(output_files,outputs,output_indices):
                o_file[out_ids]=np.asarray(out)
                del(o_file)
                #        log.info('{} elements are in queue'.format(queue.qsize())
    return mpFunction


def get_mp_function_SharedArray_multi_new(function,arguments,output_names,output_shapes,args=[],callWithMultipleArguments=True,output_file = False,output_indices=slice(None),**kwargs):
    repeatedArguments=[repeat(arg) for arg in args]
    #log.info('arguments={}'.format(arguments))
    if len(arguments)>0:
        nParts=(len(arguments[0]),)
    else:
        nParts=(1,)
        arguments=['mock_argument']
    #log.info('nParts={}'.format(nParts))
    #    log.info('newShape={}'.format(newShape))
    #log.info('repeatedArguments={}'.format(repeatedArguments))
    #log.info('call with multiple args={}'.format(callWithMultipleArguments))
    kwargs['output_ids']=output_indices
    if callWithMultipleArguments:        
        def mpFunction():            
            output_files = [sa.attach('shm://'+ name).reshape(shape) for name,shape in zip(output_names,output_shapes)]
            kwargs['outputs']=output_files
            # in this function output have to be placed in outputs[i][output_ids]
            function(*arguments,*args,**kwargs)            
    else:
        def mpFunction():
            output_files = [sa.attach('shm://'+ name).reshape(shape) for name,shape in zip(output_names,output_shapes)]
            kwargs['outputs']=output_files
            for argument,out_ids in zip(arguments,output_indices):
                kwargs['output_ids']=out_ids
                # in this function output have to be placed in outputs[i][output_ids]
                function(argument,*args,**kwargs)
    return mpFunction

###############
#####recipies#####
def id_operator(x):
    return x
class RecipeFactory:
    def __init__(self,operatorDict):
        # operatorDict is a dictionary containing Function names as keys and the function themselfs as values
        self.operatorDict={}
        self.addOperators({'id':id_operator})
        self.addOperators(operatorDict)
        self.number_of_arguments_per_operator={}
    def copy(self):
        return RecipeFactory(self.operatorDict.copy())
    
    def addOperators(self,operatorDict):
        for key,value in operatorDict.items():
            if isinstance(value,list):
                assert isinstance(value[1],int) and (len(value)==2),'opperator list has be of type [function,int:n_umber_of_arguments] but is {}'.format([type(x)for x in value])
                self.operatorDict[key]=value[0]
                self.number_of_arguments_per_operator[key]=value[1]
            else:
                self.operatorDict[key]=value

    def get_operator(self,name):
        return self.operatorDict[name]
    def buildRecipe(self,recipeSketch):
        keys={'preProcessing','processData','processResults','postProcessing'}
        recipePlan=self.parseRecipeSketch(recipeSketch,keys)
#        log.info('Parsed the Sketch to: \n {0}\n'.format(recipePlan['preProcessing']))
        recipeData={}
        for key in keys:
            processPlan=recipePlan[key]
            recipeData[key]=self.buildProcess(processPlan)

        return recipeData
    def buildProcessFromSketch(self,processSketch):
        processPlan=self.parseProcessSketch(processSketch)
        process=self.buildProcess(processPlan)
        return process
    def buildProcess(self,processPlan):
        stepList=[]
        outputInputMappingsList=[]
        previousStepPlan=False
        for outputInputMapping,stepPlan in processPlan:
            try:
                step=self.buildStep(stepPlan)
            except Exception as e:
                log.error('error in step\n {}'.format(stepPlan))
                raise e
            if outputInputMapping.shape==(0,):
                isFirstStep=isinstance(previousStepPlan,bool)
                if isFirstStep:
                    previousStepPlan=np.arange(step.numberOfInputs)
                outputInputMapping=self.getStandardMapping(previousStepPlan,step.numberOfFreeInputs)
            stepList+=[step]
            outputInputMappingsList+=[outputInputMapping]
            previousStepPlan=stepPlan
        outputInputMappingsList=tuple(map(tuple,outputInputMappingsList))
        process=Process(stepList,outputInputMappingsList)
        return process

    def buildStep(self,stepPlan):
        #requires Python 3.7 and higher ! otherwise the order of keys of a dict may not be preserved.
        stepOperatorDict={}
        operatorList=[]
        numberOfInputsPerOperatorList=[]
        
        fixedInputsStep=np.array([],dtype=object)
        fixedInputsIndexStep=np.array([],dtype=int)
        NrInputsStep=0
        for operatorSpecifier in stepPlan:
            operatorName=operatorSpecifier[0]
    #        log.info('operatorName={}'.format(operatorName))
            try:
                operator=self.operatorDict[operatorName]
            except KeyError as error:
                log.error('Operator "{}" not found in operator dictionary keys.\n Operator dictionary conthaines the following keys \n {}'.format(operatorName,self.operatorDict.keys()))
                raise
            #            log.info('operatorSpecifier[1] = {0}'.format(operatorSpecifier[1]))
            try:
                fixedInputs=operatorSpecifier[1][1]
            except IndexError as e:
                log.error(operatorSpecifier)
                raise e
            fixedInputsIndex=operatorSpecifier[1][0]+NrInputsStep
            
            fixedInputsIndexStep=np.concatenate((fixedInputsIndexStep,fixedInputsIndex))
            fixedInputsStep=np.concatenate((fixedInputsStep,fixedInputs))            

            if isinstance(operator,Process):
                NrInputs = len(operator.outputInputMappings[0])
                #log.info('Nr Inputs in initalization = {}'.format(NrInputs))
                operatorFunction = operator.run
            else:
                NrInputs=self.number_of_arguments_per_operator.get(operatorName,'no number of arguments given.')
                operatorFunction = operator                
                if not isinstance(NrInputs,int):
                    NrInputs=len(inspect.signature(operatorFunction).parameters)
            NrFreeInputs=NrInputs-operatorSpecifier[1].shape[-1]
            NrInputsStep+=NrInputs
            operatorPositionsInList=stepOperatorDict.get(operatorName,[])            
            stepOperatorDict[operatorName]=operatorPositionsInList+[len(operatorList)]            
            operatorList+= [operatorFunction]
            numberOfInputsPerOperatorList += [NrInputs]


        freeInputIndexStep=np.delete(np.arange(NrInputsStep),fixedInputsIndexStep.astype(int))
        inversInputIndexStep=np.concatenate((fixedInputsIndexStep,freeInputIndexStep)).astype(int)
        inputIndexStep=invertNumpyIndex(inversInputIndexStep) # calls class external function !

        numberOfInputsPerOperatorList=tuple(numberOfInputsPerOperatorList)
        fixedInputsStep=tuple(fixedInputsStep)
        inputIndexStep=tuple(inputIndexStep)                             
        inputData=[numberOfInputsPerOperatorList,fixedInputsStep,inputIndexStep]
#        log.info('inputData={}'.format(inputData))
        step=Step(stepOperatorDict,operatorList,inputData)
        return step

    
    def getStandardMapping(self,previousStepPlan,NrFreeInputsStep):
        numberOfOutputs=len(previousStepPlan)
        
        if numberOfOutputs==1:
            inOut=self.oneToAll(NrFreeInputsStep)
        else:
            if numberOfOutputs==NrFreeInputsStep:
                inOut=self.oneToOne(NrFreeInputsStep)
            elif numberOfOutputs<NrFreeInputsStep:
#                log.info('OutputNumber(nO) < InputNumber(nI) after step: \n{}\n No mapping data was specified! I will map the first output to the first nI-nO+1 inputs.All other outputs are mapped 1 to 1.'.format(previousStepPlan))
                inOut=self.manyToMany_standard(numberOfOutputs,NrFreeInputsStep)
            else:
                try:
                    raise AssertionError('Invalid process scetch!\n There are more outputs than inputs for the next step after step: \n{}'.format(previousStepPlan))
                except AssertionError as error:
                    log.error(error)
                    raise

        return inOut
    
    def oneToAll(self,NrInputs):
        inOut=np.zeros(NrInputs,dtype=int)
        return inOut
    
    def oneToOne(self,NrInputs):
        inOut=np.arange(NrInputs,dtype=int)
        return inOut

    def manyToMany_standard(self,nrOutputs,nrInputs):
        difference=nrInputs-nrOutputs
        firstToDifference=np.zeros(difference+1,dtype=int)
        restToRest=np.arange(1,nrOutputs,1,dtype=int)
        inOut=np.concatenate((firstToDifference,restToRest))
        return inOut

    def parseRecipeSketch(self,recipeSketch,keys):
        recipePlan={}
        for key in keys:
            processSketch=recipeSketch.get(key,[])
            recipePlan[key]=self.parseProcessSketch(processSketch)
        return recipePlan        

    
    def parseProcessSketch(self,processSketch):
        try:
            processPlan=[]
            placeholderIOMapping=np.array([],dtype=int)
            if isinstance(processSketch,str):
                stepSketch=processSketch
                stepPlan=self.parseStepSketch(stepSketch)
                processPlan=[[placeholderIOMapping,stepPlan]]
            elif isinstance(processSketch,list):
                for element in processSketch:
                    if isinstance(element,str):                    
                        stepPlan=self.parseStepSketch(element)
                        processPlan+=[[placeholderIOMapping,stepPlan]]
                    elif isinstance(element,list):
                        if len(element)==0:
                            log.info('process Sketch: Ignore empty step.')
                        else:
                            first_part = element[0]
                            if isinstance(first_part,(np.ndarray,tuple)):
                                try:
                                    assert len(element)==2
                                except AssertionError:
                                    log.error('Invalid process sketch. Element = {}'.format(element))
                                    raise
                                stepPlan=self.parseStepSketch(element[1])
                                processPlan+=[[np.asarray(first_part).astype(int),stepPlan]]
                            elif isinstance(first_part,str):
                                stepPlan=self.parseStepSketch(element)
                                processPlan+=[[placeholderIOMapping,stepPlan]]
                            else:
                                raise AssertionError('Invalid process sketch. Element = {}'.format(element))
                    else:
                        raise AssertionError('Invalid process sketch: sketch element must be either a string or a list')
            else:
                raise AssertionError('Invalid process sketch: sketch must be either a string or a list')
        except AssertionError as error:
            log.error(error)
            raise
  #      log.info('process Plan={}'.format(processPlan))
        return processPlan


    def parseStepSketch(self,stepSketch):
        try:
            stepPlan=[]
            placeholderFixedInputs=np.array([[],[]],dtype=object)
            if isinstance(stepSketch,str):            
                stepPlan=[[stepSketch,placeholderFixedInputs]]
            elif isinstance(stepSketch,tuple):
                if (len(stepSketch)==2):
                    if isinstance(stepSketch[1],(np.ndarray,tuple)):
                        stepPlan=[stepSketch[0],np.asarray(stepSketch[1],dtype = object)]
            elif isinstance(stepSketch,list):
                for element in stepSketch:
                    if isinstance(element,str):
                        stepPlan+=[[element,placeholderFixedInputs]]
                    elif isinstance(element,(list,tuple)):
                        errorMessage='Invalid step sketch format: given Format = {} expected Format = [str,ndarray] '.format(element)
                        assert (len(element)==2),errorMessage
                        assert isinstance(element[0],str),errorMessage
                        assert isinstance(element[1],(tuple,np.ndarray)),errorMessage
                        fixedInputsSketch=np.asarray(element[1],dtype = object)
                        sketchShape=fixedInputsSketch.shape
                        if (len(sketchShape)==1):
                            mapping=np.arange(sketchShape[0])
                            fixedInputsPlan=np.array([mapping,fixedInputsSketch],dtype=object)
                            element=[element[0],fixedInputsPlan]
                        else:
                            element = [element[0],fixedInputsSketch]                            
                        stepPlan += [element]
                    else:
                        raise AssertionError('Invalid step sketch format: Each entry of a sketch must be either a String or a List')
            else:
                raise AssertionError('Invalid step sketch format: sketch must be either a String or a List')
        except AssertionError as error:
            log.error(error)
            raise
        return stepPlan
    
class Process:
    def __init__(self,steps,outputInputMapping):
        def readOperatorsFromSteps(steps):
            operatorDict={}
            for stepIndex,step in enumerate(steps):
                for operatorName in step.operatorDict.keys():
                    stepList=operatorDict.get(operatorName,[])
                    operatorDict[operatorName]=stepList+[stepIndex]
            return operatorDict
            
        self.steps=steps
        self.outputInputMappings=outputInputMapping
        self.processList=tuple(zip(steps,outputInputMapping))
        self.nextInput=[]
        self.operatorInStepDict=readOperatorsFromSteps(steps)

    def changeOperator(self,operatorName,operator):
        stepIndexList=self.operatorInStepDict.get(operatorName,[])
        for index in stepIndexList:
            self.steps[index].changeOperator(operatorName,operator)
            
    def run(self,*args):
#        if not isinstance(initial_arg,tuple):
        nextInput=tuple(args)
            
#        log.info('nextInput={}'.format(nextInput))        
        processList=self.processList
        for step,mapping in processList:
            try:
                #log.info('input length={}'.format(len(nextInput)))
                #log.info('input output mapping={}'.format(mapping))
                #log.info('step ={}'.format(tuple(step.operatorDict.keys())))
                _input=tuple(map(nextInput.__getitem__,mapping))
                #log.info('alive')
                #log.info('_input length={}'.format(len(_input)))
                nextInput=step.run(_input)
                #log.info('alive2')
                #            log.info('next Input={}'.format(nextInput))
            except IndexError as e:
                log.error('Error in Process containing {}'.format(self.operatorInStepDict))
                log.error(e)
                raise
        if len(nextInput)==1:
            return nextInput[0]
        else:
            return nextInput
                    
class Step:
    def __init__(self,operatorDict,operatorList,inputData):
        self.operatorDict=operatorDict
        self.operatorList=operatorList
        self.numberOfInputsPerOperatorList=inputData[0]
        self.predefinedInputs=inputData[1]
        self.inputMapping=inputData[2]
  #      log.info('operatorDict keys={}'.format(operatorDict.keys()))
        self.numberOfInputs=sum(self.numberOfInputsPerOperatorList)
        self.numberOfFreeInputs=self.numberOfInputs-len(self.predefinedInputs)
        self.stepList=tuple(zip(self.operatorList,self.numberOfInputsPerOperatorList))
#        log.info('step list=\n{}'.format(self.stepList))

    def changeOperator(self,operatorName,operator):
        stepListPositions=self.operatorDict.get(operatorName,[])
        for position in stepListPositions:
            self.operatorList[position]=operator
        self.stepList=tuple(zip(self.operatorList,self.numberOfInputsPerOperatorList))
        
    def run(self,_input):
        try:
            output=()
            _input=self.predefinedInputs+_input
            #log.info('input={}'.format(_input))        
            #log.info('length of _input={}'.format(len(_input)))
            #log.info('input Mapping={}'.format(self.inputMapping))
            getitem=_input.__getitem__
            #        log.info('second Item={}'.format(getitem(self.inputMapping[-1])))
            _input=tuple(map(getitem,self.inputMapping))
        except Exception as e:
            log.error('Input output mismatch in step containing the operators: {}.\n Input = {} InputMapping = {}'.format(self.operatorDict.keys(),_input,self.inputMapping))
            raise e
        log.debug('running step with Operators ={}'.format(self.operatorDict.keys()))
        for operator,NrOfInputs in self.stepList:
            try:
                #log.info('operator={}'.format(operator))
                #log.info('nrOfInputs = {}'.format(NrOfInputs))
                #log.info('nrOf  available Inputs = {}'.format(len(_input)))                
                out=operator(*_input[:NrOfInputs])
                #            log.info('output={}'.format(out))
                if type(out)!=type(None):
                    output+=(out,)
                _input=_input[NrOfInputs:]
            except Exception as e:
                log.error('Error in operator {} in the step containing {}'.format(operator,self.operatorDict.keys()))
                raise e
        return output
    

########################
###### generall classes

class DictNamespace(SimpleNamespace):
    @staticmethod
    def dict_to_dictnamespace(d):
        n=DictNamespace()
        for key,value in d.items():
            if isinstance(value,dict):
                value = DictNamespace.dict_to_dictnamespace(value)
            setattr(n,str(key),value)
        return n
    @staticmethod
    def dictnamespace_to_dict(d):
        n={}
        for key,value in d.items():
            if isinstance(value,DictNamespace):
                value = DictNamespace.dictnamespace_to_dict(value)
            n[str(key)]=value
        return n
    def __init__(self,**kwargs):
        super().__init__(**kwargs)        
    def decorator_convert_dict_to_dictnamespace(func):
        def new_func(self,*args):
            if isinstance(args[-1],dict):
                args = args[:-1] + (self.dict_to_dictnamespace(args[-1]),)
            return func(self,*args)
        return new_func
    def dict(self):
        return self.dictnamespace_to_dict(self)
    def items(self):
        for key,value in self.__dict__.items():
            yield (key,value)
    def keys(self):
        for key in self.__dict__.keys():
            yield key
    def pop(self,key):
        return self.__dict__.pop(key)
    def values(self):
        for values in self.__dict__.values():
            yield values
    def __len__(self):
        return self.__dict__.__len__()
    def __iter__(self):
        return self.keys()
    def __getitem__(self, item):
        return self.__dict__[item]
    def copy(self):
        return self.dict_to_dictnamespace(self.dict())
    def get(self,*args,**kwargs):
        return self.__dict__.get(*args,**kwargs)
    
    @decorator_convert_dict_to_dictnamespace
    def __setitem__(self, key,value):
        self.__setattr__(key, value)
        
    @decorator_convert_dict_to_dictnamespace
    def update(self,data):
        self.__dict__.update(data)
            
    
    def  __getattribute__(self,key):
        try:
            value = super().__getattribute__(key)
        except AttributeError as e:
            log.error(e)
            log.info('Known attributes are {}'.format(list(self.keys())))
            raise
        return value
    
class DictNamespace2(SimpleNamespace):
    @staticmethod
    def dict_to_dictnamespace(d):
        n=DictNamespace()
        for key,value in d.items():
            if isinstance(value,dict):
                value = DictNamespace.dict_to_dictnamespace(value)
            setattr(n,str(key),value)
        return n
    @staticmethod
    def dictnamespace_to_dict(d):
        n={}
        for key,value in d.items():
            if isinstance(value,DictNamespace):
                value = DictNamespace.dictnamespace_to_dict(value)
            n[str(key)]=value
        return n
    def __init__(self,**kwargs):
        super().__init__(**kwargs)        
    def decorator_convert_dict_to_dictnamespace(func):
        def new_func(self,*args):
            if isinstance(args[-1],dict):
                args = args[:-1] + (self.dict_to_dictnamespace(args[-1]),)
            return func(self,*args)
        return new_func
    def dict(self):
        return self.dictnamespace_to_dict(self)
    def items(self):
        for key,value in self.__dict__.items():
            yield (key,value)
    def keys(self):
        for key in self.__dict__.keys():
            yield key
    def pop(self,key):
        return self.__dict__.pop(key)
    def values(self):
        for values in self.__dict__.values():
            yield values
    def __len__(self):
        return self.__dict__.__len__()
    def __iter__(self):
        return self.keys()
    def __getitem__(self, item):
        return self.__dict__[item]
    def copy(self):
        return self.dict_to_dictnamespace(self.dict())
    def get(self,*args,**kwargs):
        return self.__dict__.get(*args,**kwargs)
    
    @decorator_convert_dict_to_dictnamespace
    def __setitem__(self, key,value):
        self.__setattr__(key, value)
        
    @decorator_convert_dict_to_dictnamespace
    def update(self,data):
        self.__dict__.update(data)
            
    
    def  __getattribute__(self,key):
        try:
            value = super().__getattribute__(key)
        except AttributeError as e:
            log.error(e)
            log.info('Known attributes are {}'.format(list(self.keys())))
            raise
        return value
    
class FTGridPair:
    def __init__(self,realGrid,reciprocalGrid):
        self.realGrid=realGrid
        self.reciprocalGrid=reciprocalGrid




def option_switch(opt,default_value=True):
    '''returns True when opt is True or non-bool, if non_bool retun opt else return default value as second parameter'''
    if isinstance(opt,bool):
        if opt:
            return (True,default_value)
        else:
            return (False,default_value)
    else:
        return (True,opt)



def grow_mask(mask,n_pixels):
    mansk = np.array(mask)
    for n in range(n_pixels):        
        grads = np.sum(np.abs(np.gradient(mask.astype(int),edge_order = 1)[1:]),axis = 0)
        mask = mask | (grads!=0)
    return mask


def make_string_tex_conform(string):
    string = string.replace('_','\_')
    string = string.replace('&','\&')
    string = string.replace('{','\{')
    string = string.replace('}','\}')
    string = string.replace('%','\%')
    #log.info('corrected string = {}'.format(string))
    return string


def split_ids_by_unique_values(ids):
    '''
    Assumes input is sorted list of integers.
    For [0,0,0,1,1,2,3,3,3] this funciton returns [3,5,6]
    '''
    split_ids = np.nonzero(np.diff(ids))[0]+1
    return split_ids

def convert_to_slice_if_possible(frame_ids):
    #log.info('to slice = {}'.format(frame_ids))
    frame_ids = np.asarray(frame_ids)
    if len(frame_ids) ==0:
        #empty slice
        return slice(1,0)
    elif len(frame_ids) == 1:
        start = frame_ids[0]
        return slice(start,start+1)
    else:
        start = frame_ids[0]
        step = frame_ids[1] - start
        stop = frame_ids[-1] + step
        #log.info('start = {}, stop = {} step = {}'.format(start,stop,step))
        #log.info('frame_ids = {}'.format(frame_ids))
        sliced_ids = np.arange(start,stop,step,dtype = frame_ids.dtype)
        if len(frame_ids) == len(sliced_ids):
            if (frame_ids == sliced_ids).all():            
                #log.info('frame selection is slice like!')
                frame_ids = slice(start,stop,step)                    
    return frame_ids

def split_into_simple_slices(sequence,return_length=False,return_sliced_args=False,mod=False):
    '''
    splits a 1D non decreasing list of integers into slices which correspond to
    1. if mod = False
    its connected components.  Where two integers i,j are connected if |j-i|=1, eg if they are consequtive.
    2. if mod = int:  The same but in equivalence classes of Z/modZ
    '''
    slices = []
    arg_slices = []
    lengths = []
    if isinstance(mod,int) and (not isinstance(mod,bool)):
        eq_class_ids = sequence%mod
        for c_id in range(mod):
            eq_class = sequence[eq_class_ids==c_id]
            if len(eq_class)<=0:
                continue
            eq_min = eq_class[0]
            temp = (eq_class-eq_min)//mod
            jumps = np.nonzero(np.diff(temp)!=1)[0]+1
            connected_components = np.split(eq_class,jumps)
            #connecteda_args = np.split(np.arange(len(sequence)),jumps)
            for c in connected_components:
                if len(c) >0:
                    slices.append(slice(c[0],c[-1]+mod/2,mod))
                    arg_slices.append(slice(c[0]-eq_min,c[-1]+mod/2-eq_min,mod))
                    lengths.append(len(c))
    else:
        jumps = np.nonzero(np.diff(sequence)!=1)[0]+1
        connected_components = np.split(sequence,jumps)
        connected_args = np.split(np.arange(len(sequence)),jumps)
        _min=sequence.min()                                
        for c,ca in zip(connected_components,connected_args):
            if len(c) >0:
                slices.append(slice(c[0],c[-1]+1))
                arg_slices.append(slice(ca[0],ca[-1]+1))
                lengths.append(len(c))
            
    if return_length and return_sliced_args:
        return slices,arg_slices,lengths
    elif return_length:
        return slices,lengths
    elif return_sliced_args:
        return slices,arg_slices
    else:
        return slices    
    
###### cache aware routines ######
def get_L2_cache_split_parameters(data_shape,data_type,L2_cache):
    unit_size = data_type.itemsize
    units_in_L2 =  L2_cache*1024/unit_size
    submatrix_sizes = np.array([np.prod(np.array(data_shape[i:])) for i in range(len(data_shape))])
    #log.info('submatrix_sizes = {}'.format(submatrix_sizes))
    #log.info("units in  L2 + {}".format(units_in_L2))
    if submatrix_sizes[0]>units_in_L2:
        splitting_dimension = np.nonzero(submatrix_sizes//units_in_L2)[0][-1]
    else:
        splitting_dimension = -1

    #log.info('unit size {} units in L2 {} data shape {} submatrix_sizes {} splitting_dimension = {}'.format(unit_size,units_in_L2,data_shape,submatrix_sizes,splitting_dimension))
    
    if splitting_dimension >= len(data_shape)-1:
        fixed_sub_size = 1*unit_size
    else:
        fixed_sub_size = submatrix_sizes[splitting_dimension+1]
        
    n_sub_sizes_in_L2 = int(units_in_L2//fixed_sub_size)    
    step = n_sub_sizes_in_L2
    #log.info('step = {}'.format(step))
    return (splitting_dimension,step)

def _generate_conjugate_default(out_array):
    conj = np.conj
    def abs_value(data):
        return conj(data,out=out_array)
    return out_array

def _generate_conjugate_cache_aware(out_array,L2_cache):
    data_shape = out_array.shape
    data_type = out_array.dtype
    splitting_dimension,step = get_L2_cache_split_parameters(data_shape,data_type,L2_cache)

    mult = np.multiply
    sqrt = np.sqrt
    conj = np.conjugate
    def conj_1_loop(data):
        for i in range(0,data_shape[0],step):
            i2 = i+step
            d = data[i:i2]
            o = out_array[i:i2]
            conj(d,out = o)
        return out_array
    def conj_2_loop(data):
        for i in range(data_shape[0]):
            for j in range(0,data_shape[1],step):
                j2 = j+step
                d = data[i,j:j2]
                o = out_array[i,j:j2]
                conj(d,out = o)
        return out_array
    def conj_3_loop(data):
        for i in range(data_shape[0]):
            for j in range(data_shape[1]):
                for k in range(0,data_shape[2],step):
                    k2 = k+step
                    d = data[i,j,k:k2]
                    o = out_array[k:k2]
                    conj(d,out = o)
        return out_array
    
    if splitting_dimension == -1:
        abs_value = _generate_conjugate_default(out_array)
    elif splitting_dimension == 0:
        abs_value = conj_1_loop
    elif splitting_dimension == 1:
        abs_value = conj_2_loop
    elif splitting_dimension == 2:
        abs_value = conj_3_loop
    return abs_value

def generate_conjugate(data_shape,data_type,cache_aware=False,L2_cache=256):
    out = np.zeros(data_shape,dtype = data_type)
    if not cache_aware:
        abs_value = _generate_conjugate_default(out)
    else:
        abs_value = _generate_conjugate_cache_aware(out,L2_cache)
    return abs_value


#### Testing routines ####
def default_test(result,expected,precision=False, name='default'):
    #if isinstance(result,np.ndarray) and isinstance(expected,np.ndarray):
    #    result = [result]
    #    expected = [expected]
    def listsAreEqual(result,expected):
        areEqual=True
        if len(result)==len(expected):
            for resPart,expPart in zip(result,expected):
                if isinstance(resPart,(list,tuple)) and isinstance(expPart,(list,tuple)):
                    areEqual = areEqual and listsAreEqual(resPart,expPart)
                elif isinstance(resPart,np.ndarray) and isinstance(expPart,np.ndarray):
                    if resPart.dtype==np.object and expPart.dtype==object and len(resPart)>1:
                        areEqual = areEqual and listsAreEqual(resPart,expPart)
                    else:
                        if precision==False:
                            areEqual = areEqual and (resPart==expPart).all()
                        else:
                            areEquaL=areEqual and np.allclose(resPart,expPart,atol=precision)
                else:
                    try:
                        areEqual = areEqual and (resPart == expPart)
                    except Exception as e:
                        areEqual = False
                if not areEqual:
                    #log.info('{},{}'.format(result,expected))
                    break
        else:
            areEqual=False
        return areEqual

    test_successful = False
    if not (isIterable(result) or isIterable(expected)):
        try:
            test_successful = (result == expected)
        except Exception as e:
            pass
    else:
        test_successful = listsAreEqual(result,expected)
    try:
        assert test_successful
        print('Test {}: Passed!'.format(name))
    except AssertionError:
        print('Test {} Failed: \n Expected:{} \n Got:{}'.format(name,expected,result))
        raise


#### process threshold ####
def create_threshold_projection(threshold):
    no_lower_threshold  = (not isinstance(threshold[0],(float,int))) or isinstance(threshold[0],bool)
    no_upper_threshold = (not isinstance(threshold[1],(float,int))) or isinstance(threshold[1],bool)
    #log.info('no upper thresh = {} now lower thresh = {}'.format(no_upper_threshold,no_lower_threshold))
    if no_upper_threshold and (not no_lower_threshold):
        def threshold_projection(density):
            real_part = density.real            
            is_invalid_mask = real_part<threshold[0]
            #log.info('n = {} pixels lower than {}'.format(is_invalid_mask.sum(),threshold[0]))
            real_part[is_invalid_mask] = threshold[0]           
            return [density,is_invalid_mask]
    elif (not no_upper_threshold) and no_lower_threshold:
        def threshold_projection(density):
            real_part = density.real
            is_invalid_mask = real_part>threshold[1]
            real_part[is_invalid_mask] = threshold[1]            
            return [density,is_invalid_mask]
    elif no_upper_threshold and no_lower_threshold:
        def threshold_projection(density):
            return [density,False]
    else:
        def threshold_projection(density):
            real_part = density.real
            is_smaller_mask = (real_part<threshold[0])
            is_bigger_mask = (real_part>threshold[1])
            real_part[is_smaller_mask] = threshold[0]
            real_part[is_bigger_mask] = threshold[1]            
            return [density,is_bigger_mask | is_smaller_mask] 
    return threshold_projection


#### hide print ####
from contextlib import contextmanager

@contextmanager
def stdout_redirected(to=os.devnull):
    '''
    import os

    with stdout_redirected(to=filename):
        print("from Python")
        os.system("echo non-Python applications are also supported")
    '''
    fd = sys.stdout.fileno()

    ##### assert that Python and C stdio write using the same file descriptor
    ####assert libc.fileno(ctypes.c_void_p.in_dll(libc, "stdout")) == fd == 1

    def _redirect_stdout(to):
        sys.stdout.close() # + implicit flush()
        os.dup2(to.fileno(), fd) # fd writes to 'to' file
        sys.stdout = os.fdopen(fd, 'w') # Python writes to fd

    with os.fdopen(os.dup(fd), 'w') as old_stdout:
        with open(to, 'w') as file:
            _redirect_stdout(to=file)
        try:
            yield # allow code to be run with the redirected stdout
        finally:
            _redirect_stdout(to=old_stdout) # restore stdout.
                                            # buffering and flags such as
                                            # CLOEXEC may be different
def stdout_redirect(to):
    fd = sys.stdout.fileno()
    sys.stdout.close() # + implicit flush()
    os.dup2(to.fileno(), fd) # fd writes to 'to' file
    sys.stdout = os.fdopen(fd, 'w') # Python writes to fd
        



import os
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1" 
os.environ['OPENBLAS_NUM_THREADS'] = "1"


import numpy as np
import scipy
import logging
import traceback
from xframe import log
os.chdir(os.path.expanduser('~/Programs/xframe'))
from xframe.startup_routines import load_recipes
from xframe.startup_routines import dependency_injection_no_soft
from xframe.library import mathLibrary
from xframe.externalLibraries.flt_plugin import LegendreTransform
mathLibrary.leg_trf = LegendreTransform
from xframe.externalLibraries.shtns_plugin import sh
mathLibrary.shtns = sh
from xframe.control.Control import Controller
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_to_deg2_invariant
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_mask
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_apply_precision_filter
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_to_deg2_invariant_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import bl_3d_pseudo_inverse_worker
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_cc_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import density_to_deg2_invariants_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_eigenvalues
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_legendre_matrices
from xframe.plugins.MTIP.analysisLibrary import fxs_invariant_tools as i_tools
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_weightDict_zernike_spherical
from xframe.plugins.MTIP.analysisLibrary.harmonic_transforms import HarmonicTransform
from xframe.plugins.MTIP.analysisLibrary.fourier_transforms import generate_zernike_spherical_ft
from xframe.plugins.MTIP.analysisLibrary.ft_grid_pairs import get_grid
from xframe.plugins.MTIP.analysisLibrary.fxs_Projections import ReciprocalProjection_old,ReciprocalProjection
from xframe import settings
from xframe.library import mathLibrary as mLib
from xframe import Multiprocessing
from xframe.library import physicsLibrary as pLib
from xframe.library.gridLibrary import GridFactory
log=log.setup_custom_logger('root','INFO')

dependency_injection_no_soft()
analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_pentagon_100')[0]
db = analysis_sketch[1]
opt = settings.analysis
controller = Controller(analysis_sketch)
from xframe.presenters.matplolibPresenter import heat2D,heat2D_multi


def init_transforms(maxQ):
    max_order = 99
    n_radial_points = 128
    opt=settings.analysis
    ht_opt={'dimensions':3,'max_order':max_order,'indices':'lm','anti_aliazing_degree':2,'n_phi':0,'n_theta':0}
    #ht_opt={'dimensions':opt.dimensions,**opt['grid'],'type':opt['fourier_transform']['type'],'pi_in_q':opt['fourier_transform']['pi_in_q'],'reciprocal_cut_off':maxQ}
    # harmonic transforms and grid
    cht=HarmonicTransform('complex',ht_opt)
    weight_dict = generate_weightDict_zernike_spherical(max_order,n_radial_points)
    grid_opt={'dimensions':3,'type':'Zernike_dict','max_q':maxQ,'n_radial_points':n_radial_points,**cht.grid_param,'pi_in_q':True}
    grid_pair = get_grid(grid_opt)
    maxR = grid_pair.real[:,0,0].max()
    ft,ift = generate_zernike_spherical_ft(maxR,weight_dict,cht,'complex',True,use_gpu=True)
    orders = np.arange(max_order+1)
    return locals()


ccd = db.load('ccd')

#mask_dict_true={'type':'direct','direct':{'mask':np.full(cc_1.shape,True,dtype = bool)}}
#bl,_mask = cross_correlation_to_deg2_invariant(ccd['cross_correlation'],3,{**ccd,'orders':orders,'mask':mask_dict_true,'mode':'pseudo_inverse'})

phis = ccd['phis']
phi_mask = phis<np.pi
phis = phis[phi_mask]
qs = ccd['qs']
thetas = ccd['thetas']
orders = np.arange(0,50)
q_ids = np.arange(128)

leg_m = ccd_legendre_matrices(q_ids,q_ids,phis,thetas,orders)
leg_m_inv = np.linalg.pinv(leg_m)
bl_initial = bl[:,q_ids,q_ids]


cc_initial = np.sum(bl_initial[...,None]*np.moveaxis(leg_m,-1,0),axis=0)


n_m_phi = 200
leg_m = ccd_legendre_matrices(q_ids,q_ids,phis[n_m_phi:-n_m_phi],thetas,orders)
leg_m_inv = np.linalg.pinv(leg_m)
bl_recovered = np.sum(cc_initial[:,None,n_m_phi:-n_m_phi]*leg_m_inv,axis=-1)

cc_mask = np.full((len(qs),len(qs),len(phis)),True)
cc_mask[...,:n_m_phi]=False
cc_mask[...,-n_m_phi:]=False
cc_compatible = np.zeros(cc_mask.shape,dtype=float)
cc_compatible[q_ids,q_ids]=cc_initial

bl_recovered_2 = bl_3d_pseudo_inverse_worker(q_ids,q_ids,cc_compatible,phis,thetas,orders,cc_mask)

diff = np.abs(bl_recovered[1:]-bl_initial.T[1:])
#diff = np.abs(bl_recovered_2[1:]-bl_recovered[1:])
error = diff/np.abs(bl_initial.T[1:])
#error = diff/np.abs(bl_recovered[1:])
error[diff==0] = 0
print(error[-1,:])
print(error.max())




import numpy as np
import logging

log=logging.getLogger('root')

from xframe.library import mathLibrary as mLib
from xframe.library.mathLibrary import PolarIntegrator,SphericalIntegrator
from xframe.library.mathLibrary import gaussian_fourier_transformed_spherical
from scipy.optimize import minimize_scalar

class ShrinkWrapParts:
    def __init__(self,initial_support,real_grid,reciprocal_grid,threshold,gaussian_sigma,mode = 'threshold',mode_options = {}):
        dimension = real_grid[:].shape[-1]
        self.mode_routines = {'threshold':self.generate_get_new_mask_threshold,'fixed_volume':self.generate_get_new_mask_fixed_volume}
        self.mode_options
        self.mode = mode
        self.real_grid = real_grid[:]
        self.reciprocal_grid = reciprocal_grid[:]
        dimension = self.real_grid.shape[-1]
        log.info(f'SW dim = {dimension}')
        if dimension ==2:
            self.integrator = PolarIntegrator(self.real_grid)            
            self._default_gaussian_sigma = real_grid[1,0,0]-real_grid[0,0,0]
        elif dimension == 3:
            self.integrator = SphericalIntegrator(self.real_grid)
            self._default_gaussian_sigma = real_grid[1,0,0,0]-real_grid[0,0,0,0]
        self.initial_volume = self.integrator.integrate_normed(initial_support.astype(float))
        
        self._threshold = [threshold]
        
        self._gaussian_sigma = [gaussian_sigma]
        self.gaussian_values = gaussian_fourier_transformed_spherical(self.grid,self._gaussian_sigma[0])
        
        self.get_new_mask = self.mode_routines.get[self.mode]()
        self.multipy_with_ft_gaussian = self.generate_multiply_by_ft_gaussian()
    @property
    def threshold(self):
        return self._threshold[0]
    @threshold.setter
    def threshold(self,value):
        if value<=0:
            self._threshold[0]=0
            log.warning('Shrikwrap threshold has to lie in [0,1] but given value is {}. Projecting threshold to {}.'.format(value,self._threshold[0]))
        elif value>=1:
            self._threshold[0]=1
            log.warning('Shrikwrap threshold has to lie in [0,1] but given value is {}. Projecting threshold to {}.'.format(value,self._threshold[0]))
        else:
            self._threshold[0]=value


    @property
    def gaussian_sigma(self):
        return self._gaussian_sigma[0]
    @gaussian_sigma.setter
    def gaussian_sigma(self,value):
        if value>=0:
            self._gaussian_sigma[0]=value
        else:
            self._gaussian_sigma[0]=0
            log.warning('Gaussian sigma has to be grater than 0 but given value is {}. Projecting threshold to {}.'.format(value,self._gaussian_sigma[0]))
        self.gaussian_values[:] = gaussian_fourier_transformed_spherical(self.grid,self._gaussian_sigma[0])
        
    def generate_get_new_mask_threshold(self):
        threshold = self.threshold
        def get_new_mask(convolution_data):
            abs_array=convolution_data
            max_value= abs_array.max()
            min_value= abs_array.min()
            diff = max_value-min_value
            new_mask = abs_array >= min_value + threshold[0]*diff
            return new_mask
        return update_mask
    
    def generate_get_new_mask_fixed_volume(self):
        target_volume = self.initial_volume*self.mode_options['volume'] # number in [0,1] indication volume fraction relative to self.initial_volume
        integrate = self.integrator.integrate_normed
        def get_new_mask(convolution_data):
            abs_array=convolution_data
            max_value= abs_array.max()
            min_value= abs_array.min()
            diff = max_value-min_value
            def new_volume(threshold):                
                new_mask = abs_array >= min_value + threshold*diff
                new_volume = abs(integrate(new_mask.astype(float))-target_volume)
            opti_result = minimize_scalar(new_volume,bounds=[0,1])
            threshold = opti_result.x
            log.info('Optimization result = {}'.format(opti_result.message))
            new_mask = abs_array >= min_value + threshold*diff
            return new_mask
        return update_mask

    
    def generate_multiply_by_ft_gaussian(self):
        gaussian_values = self.gaussian_values
        def multiply_ft_gaussian(data):
            return data*gaussian_values
        return multipy_with_ft_gaussian


    
def generateSW_SupportMaskRoutine(threshold):
    def generateSupportMask(convolutionData):
        absArray=np.abs(convolutionData)
        maxValue=np.max(absArray)
#        log.info('max value convoluted data={} max data value={}'.format(maxValue,np.max(absArray)))
        supportMask=absArray>=threshold*maxValue
#        positivity_mask=real_grid>=0
#        log.info('support mask={}'.format(supportMask))
        return supportMask
    return generateSupportMask

def generateSW_multiply_ft_gaussian(sigma,grid):
    dim=grid.total_shape[-1]
    #reciprocal grid needed
    if dim == 2:        
        gaussianArray=gaussian_fourier_transformed_spherical(grid,sigma)
    elif dim == 3:
        gaussianArray=gaussian_fourier_transformed_spherical(grid,sigma)
        
    else:
        e=AssertionError('multiply_ft_gaussion currently only implemented for dim=2,3')
        log.error(e)
        raise e
    
    def multiply_ft_gaussian(data):
        #            pres.present(gaussianArray,grid=ft_grid_pair.realGrid,layout={'title':'gaussian to multiply'})
        return data*gaussianArray
    return multiply_ft_gaussian



def SW_initialSupportMask(fxsData,threshold,realGrid,constantMaxR=False):
    if not isinstance(constantMaxR,bool):
        log.info('constantMaxR={}'.format(constantMaxR))
        supportMask=np.where(realGrid[:][:,0]<constantMaxR,True,False).reshape(realGrid.shape)
    else:
        maxValue=np.max(fxsData.ACD)
        #log.info('max value convoluted data={} max data value={}'.format(maxValue,np.max(absArray)))
        supportMask=absArray>=threshold*maxValue
    log.info('support mask[:]={}'.format(supportMask))
#    pres=heatPolar2D()
#    pres.present(realGrid,supportMask)
    return supportMask



import logging
import numpy as np


from xframe.interfaces import DetectorInterface
from xframe.library.mathLibrary import plane3D
from xframe.library.gridLibrary import GridFactory

log=logging.getLogger('root')

class AGIPD(DetectorInterface):
    dimensions=3
    number_of_modules=16    
    groups = np.array([[[12,13,14,15],[8,9,10,11]],[[0,1,2,3],[4,5,6,7]]],dtype = int)
    modules_per_group=4
    module_width_in_pixel = 512+7
    module_height_in_pixel = 128
    data_shape = (16,512,128)
    sensitive_pixel_mask = np.full((16,519,128),True)
    framed_sensitive_pixel_mask = np.full((16,521,130),False)
    sensitive_pixel_mask[:,64:][:,::65]=False
    framed_sensitive_pixel_mask[:,1:-1,1:-1]=sensitive_pixel_mask
    asic_slices = [
        [
            [slice(i * 64, i*64+64),slice(0,64)],
            [slice(i * 64, i*64+64),slice(64,128)],
        ]
        for i in range(8)]
    
    def __init__(self,database,load_geometry_file=False,**kwargs):
        self.database=database                
        self._origin = np.zeros(3,dtype = float)
        self.quadrants=np.zeros([2,2])
        self.pixel_grid=np.zeros([self.number_of_modules,self.module_width_in_pixel+1, self.module_height_in_pixel+1,self.dimensions])
        self.framed_pixel_grid=np.zeros([self.number_of_modules,self.module_width_in_pixel+1+2, self.module_height_in_pixel+1+2,self.dimensions])
        self.framed_pixel_centers=np.zeros([self.number_of_modules,self.module_width_in_pixel+2, self.module_height_in_pixel+2,self.dimensions])
        
        modules=[]
        modules.append(AGIPDmodule(0))
        for id in np.arange(1,self.number_of_modules,1):
            newModule=AGIPDmodule(id,pixel_grid=modules[0].pixel_grid)
            modules.append(newModule)
        modules=np.array(modules)
        self.modules=modules
        
        quadrant0=AGIPDmoduleGroup(modules[self.groups[0,0]])
        quadrant1=AGIPDmoduleGroup(modules[self.groups[0,1]])
        quadrant2=AGIPDmoduleGroup(modules[self.groups[1,0]])
        quadrant3=AGIPDmoduleGroup(modules[self.groups[1,1]])
        self.quadrants=np.array([[quadrant0,quadrant2],[quadrant1,quadrant3]])
        self.assemblePixelGrid()
        if load_geometry_file:
            self.loadGeometryFile()
        self._pixel_corner_index = False
        

    @property
    def origin(self):
        return self._origin
    @origin.setter
    def origin(self,origin:np.ndarray):
        transitition_vect = origin - self._origin 
        for module in self.modules:
            plane = module.detection_plane
            plane.base = plane.base + transitition_vect
            module.detection_plane = plane
        self.assemblePixelGrid()
        self._origin = origin

    
    def assemblePixelGrid(self):
        for module in self.modules:            
            self.pixel_grid[module.id]=module.pixel_grid
            self.framed_pixel_grid[module.id]=module.framed_pixel_grid
            self.framed_pixel_centers[module.id]=module.framed_pixel_centers
            
    def moveModuleTo(self,moduleId,plane):
        self.modules[moduleId].detection_plane = plane
        self.pixel_grid[moduleId] = modules[moduleId].pixel_grid
        self.framed_pixel_grid[moduleId] = modules[moduleId].framed_pixel_grid
        self.framed_pixel_centers[moduleId]=modules[moduleId].framed_pixel_centers
                    
    def loadGeometryFile(self):
        modulePlains=self.database.load('geometry')
        listOfQuadrants=self.quadrants.flatten()
        for module in self.modules:
            module.detection_plane = modulePlains[module.id]
        self.assemblePixelGrid()
                
    def get_geometry(self):
        return self.pixel_grid        
    @property
    def pixel_corner_index(self):
        if not isinstance(self._pixel_corner_index,np.ndarray):
            self._pixel_corner_index = self.generate_pixel_corner_index()
        return self._pixel_corner_index

    def generate_pixel_corner_index(self):
        pixel_grid = self.pixel_grid
        indices = np.arange(np.prod(pixel_grid.shape[:-1])).reshape(pixel_grid.shape[:-1])
        pixel_corner_index=[]
        for module in indices:
            module_indices = np.concatenate((module[:-1,1:,None],module[:-1,:-1,None],module[1:,:-1,None],module[1:,1:,None]),axis = -1)
            pixel_corner_index.append(module_indices)
        pixel_corner_index = np.array(pixel_corner_index).astype(np.int32)
        return pixel_corner_index

    
class AGIPDmoduleGroup:
    def __init__(self,modules):
        self.modules={}
        
        for module in modules:
            self.modules['{id}'.format(id=module.id)]=module
        

    def moveTo(self, quadrantPlain):
        print('Work In progress')

    
class AGIPDmodule:
    ''' 
    Each module consists of a rigid sensor of 128x512+7 pixels. Standard pixels are squares of width 0.2 mm. 
    Each 65'th column consists of insensitive pixels with doubled width, i.e. 0.4x0.2 mm (exept the last pixel row).

    For specifics about the pixel structure in each module look at the paper from Allagholi et al. (ISSN 16005775) 
    The Adaptive Gain Integrating Pixel Detector at the European XFEL.
    Section 4.2
    '''    
    _spaceDim=3
    _numberOfGainStages=3
    _numberOfMemoryCells=352
    _standardPixelSize=np.array([.2,.2]) # pixel size in millimeter 
    _widePixelSize= np.array([.4,.2]) # pixel size in millimeter
    _standardPixelSize3D=np.array([.2,.2,0]) # pixel size in millimeter 
    _widePixelSize3D= np.array([.4,.2,0]) # pixel size in millimeter
    _widePixelColumnSeparation=65
    _widthInPixel=512+7
    _heightInPixel=128
    def _construct_local_pixel_grid(_widthInPixel,_heightInPixel,_widePixelColumnSeparation,_widePixelSize,_standardPixelSize):       
        localPixelGridInPixels=np.stack(np.meshgrid(np.arange(_widthInPixel+1), np.arange(_heightInPixel+1),indexing='ij' ),2)
        widePixelContribution=np.zeros(localPixelGridInPixels.shape)
        for widePixelColumn in np.arange(_widePixelColumnSeparation,_widthInPixel,_widePixelColumnSeparation):
            widePixelContribution[widePixelColumn:,:]+=np.array([_widePixelSize[0]-_standardPixelSize[0],0])
        localPixelGrid2D=localPixelGridInPixels*_standardPixelSize+widePixelContribution
        localPixelGrid3D=np.insert(localPixelGrid2D,2,0,axis=2)        
        return localPixelGrid3D
    def _construct_local_framed_pixel_centers(local_framed_pixel_grid,_standardPixelSize3D,_widePixelSize3D,_widePixelColumnSeparation):
        local_pixel_centers =  local_framed_pixel_grid[:-1,:-1]+_standardPixelSize3D/2
        local_pixel_centers[_widePixelColumnSeparation:-2:_widePixelColumnSeparation] += (_widePixelSize3D-_standardPixelSize3D)/2
        return local_pixel_centers
    def _construct_local_framed_pixel_grid(_framedWidthInPixel,_framedHeightInPixel,_widePixelColumnSeparation,_widePixelSize,_standardPixelSize):
        localPixelGridInPixels=np.stack(np.meshgrid(np.arange(-1,_framedWidthInPixel), np.arange(-1,_framedHeightInPixel),indexing='ij' ),2)
        widePixelContribution=np.zeros(localPixelGridInPixels.shape)
        for widePixelColumn in np.arange(_widePixelColumnSeparation+1,_framedWidthInPixel,_widePixelColumnSeparation):
            widePixelContribution[widePixelColumn:,:]+=np.array([_widePixelSize[0]-_standardPixelSize[0],0])
        localPixelGrid2D=localPixelGridInPixels*_standardPixelSize+widePixelContribution
        localPixelGrid3D=np.insert(localPixelGrid2D,2,0,axis=2)
        return localPixelGrid3D

    localPixelGrid=_construct_local_pixel_grid(_widthInPixel,_heightInPixel,_widePixelColumnSeparation,_widePixelSize,_standardPixelSize)
    localFramedPixelGrid = _construct_local_framed_pixel_grid(_widthInPixel+2,_heightInPixel+2,_widePixelColumnSeparation,_widePixelSize,_standardPixelSize)
    localFramedPixelCenters = _construct_local_framed_pixel_centers(localFramedPixelGrid,_standardPixelSize3D,_widePixelSize3D,_widePixelColumnSeparation)

    def __init__(self,id,detection_plane = False, pixel_grid = False):
        self.pixel_grid=np.zeros([self._widthInPixel+1,self._heightInPixel+1,self._spaceDim])
        self.framed_pixel_grid=np.zeros([self._widthInPixel+3,self._heightInPixel+3,self._spaceDim])
        self.framed_pixel_centers=np.zeros([self._widthInPixel+2,self._heightInPixel+2,self._spaceDim])
        if isinstance(detection_plane,plane3D):
            self._detection_plane = detection_plane
        else:
            self._detection_plane = plane3D()#detection_plane
            
        self.id=id
        if isinstance(pixel_grid,bool):
            self._update_pixel_grid()
        else:
            self.pixel_grid=pixel_grid
    @property
    def detection_plane(self):
        return self._detection_plane
    @detection_plane.setter
    def detection_plane(self, plane:plane3D):
        self._detection_plane = plane
        self._update_pixel_grid()

    def _update_pixel_grid(self):
        plane = self._detection_plane
        base=plane.standardForm['base']
        x_direction=plane.standardForm['x_direction']
        y_direction=plane.standardForm['y_direction']
        
        transformationMatrix=np.array([x_direction,y_direction,np.zeros(3)]).T
        shape=self.pixel_grid.shape
        framed_shape = self.framed_pixel_grid.shape
        framed_center_shape = self.framed_pixel_centers.shape
        self.pixel_grid=base+transformationMatrix.dot(self.localPixelGrid.reshape(-1,3).T).T.reshape(shape)
        self.framed_pixel_grid=base+transformationMatrix.dot(self.localFramedPixelGrid.reshape(-1,3).T).T.reshape(framed_shape)        
        self.framed_pixel_centers=base+transformationMatrix.dot(self.localFramedPixelCenters.reshape(-1,3).T).T.reshape(framed_center_shape)       


        



import traceback
import os
import sys
import xframe
import logging
from importlib import util
log = logging.getLogger('root')

projects_to_load = xframe.settings.general.load_projects
def create_init_file_if_non_existing(project_path):
    init_path = os.path.join(project_path,'__init__.py')
    
    if not os.path.exists(init_path):
        with open(init_path, 'w') as fp:
            fp.write('')
    
def import_project(name,path):
    create_init_file_if_non_existing(path)
    py_import_path = f'xframe.projects.{name}'
    spec = util.spec_from_file_location(py_import_path, path+'__init__.py')
    module = util.module_from_spec(spec)
    sys.modules[py_import_path] = module
    spec.loader.exec_module(module)
    globals().update({name:module})

for project_name,project_path in xframe.known_projects.items():
    try:
        if isinstance(projects_to_load,(list,tuple)):
            if project_name in projects_to_load:
                import_project(project_name,project_path)
        else:        
            import_project(project_name,project_path)
    except Exception as e:
        log.error(f'Error during import of project at {project_path}, Traceback:\n'+traceback.format_exc())

        



import numpy as np
from xframe.settings import general
general.n_control_worker=0
import xframe

xframe.setup_experiment('fxs3046_online','normalized_proc')

opt = {
    'run':169,
    'data_mode': 'proc',
    'frame_range': slice(None),
    'selection': {
        'cells': slice(1,177),
        'pulses': slice(None),
        'trains': slice(0,-100,1)
                  },
    'modules':np.arange(16),
    'n_frames':20000,
    'good_cells':np.arange(1,202),
    'in_multiples_of':False,
}
exp =xframe.experiment_worker
gen = exp.get_data(opt)
geom = exp._get_geometry_rois()







from types import SimpleNamespace
import numpy as np
import traceback
import logging
log = logging.getLogger('root')

class DictNamespace(SimpleNamespace):
    @staticmethod
    def dict_to_dictnamespace(d):
        n=DictNamespace()
        for key,value in d.items():
            if isinstance(value,dict):
                value = DictNamespace.dict_to_dictnamespace(value)
            setattr(n,str(key),value)
        return n
    @staticmethod
    def dictnamespace_to_dict(d):
        n={}
        for key,value in d.items():
            if isinstance(value,DictNamespace):
                value = DictNamespace.dictnamespace_to_dict(value)
            n[str(key)]=value
        return n
    def __init__(self,**kwargs):
        super().__init__(**kwargs)        
    def decorator_convert_dict_to_dictnamespace(func):
        def new_func(self,*args):
            if isinstance(args[-1],dict):
                args = args[:-1] + (self.dict_to_dictnamespace(args[-1]),)
            return func(self,*args)
        return new_func
    def dict(self):
        return self.dictnamespace_to_dict(self)
    def items(self):
        for key,value in self.__dict__.items():
            yield (key,value)
    def keys(self):
        for key in self.__dict__.keys():
            yield key
    def pop(self,key):
        return self.__dict__.pop(key)
    def values(self):
        for values in self.__dict__.values():
            yield values
    def __len__(self):
        return self.__dict__.__len__()
    def __iter__(self):
        return self.keys()
    def __getitem__(self, item):
        return self.__dict__[item]
    def copy(self):
        return self.dict_to_dictnamespace(self.dict())
    def get(self,*args,**kwargs):
        return self.__dict__.get(*args,**kwargs)
    
    @decorator_convert_dict_to_dictnamespace
    def __setitem__(self, key,value):
        self.__setattr__(key, value)
        
    @decorator_convert_dict_to_dictnamespace
    def update(self,data):
        self.__dict__.update(data)
            
    
    def  __getattribute__(self,key):
        try:
            value = super().__getattribute__(key)
        except AttributeError as e:
            print(e)
            print('Known attributes are {}'.format(list(self.keys())))
            raise
        return value
    

class SettingsConverter:
    @classmethod
    def list_to_str(cls,l,end=']'):
        out='['
        for part in l:
            if isinstance(part,(list,tuple)):
                out+= cls.list_to_str(part,end='],')
            else:
                out+= str(part)+','            
        return out+end
    @classmethod
    def numpy_to_list(cls,array):
        out=[]
        if array.ndim==1:
            out=array.tolist()
        elif array.ndim>1:
            for part in array:
                out += [cls.numpy_to_list(part)]
        return out
    @classmethod
    def numpy_to_str(cls,array):
        dtype=array.dtype
        if dtype==np.dtype('object'):
            raise AssertionError('Cannot convert numpy array of type object to yaml settings.')
    
        l = cls.numpy_to_list(array)
        st = cls.list_to_str(l)
        numpy_str = f'np.array({st},dtype=np.{str(dtype)})'
        return numpy_str
    @classmethod
    def numpy_type_to_str(cls,val):
        _type= type(val).__name__
        str_val=str(val)
        numpy_str = f'np.{_type}({str_val})'
        return numpy_str
    @classmethod
    def recursive_convert_list(cls,l):
        out=[]
        for _id,val in enumerate(l):
            if isinstance(val,(np.number,np.bool_)):
                out.append(val.item())
            elif isinstance(val,(type(None),str,complex,float,int,bytes,bool)):
                # order is important otherwise np.float64 will be treated as float
                out.append(val)                
            elif isinstance(val,np.ndarray):
                log.warning(f'Numpy arrays not allowed in lists or tuples when saving settings to yaml. Skipping array.')                
            elif isinstance(val,(list,tuple)):                
                out.append(cls.recursive_convert_list(val))
            else:
                log.warning(f'Could not convert element {val} of type {type(val)} to yaml settings format.Skipping.')
        return out
                                        
    @classmethod
    def recoursive_convert_settings(cls,settings):
        out_dict={}
        for key,val in settings.items():
            key = str(key)
            try:
                if isinstance(val,(type(None),str,complex,float,int,bytes,bool)):
                    out_dict[key]=val
                elif isinstance(val,(np.number,np.bool_)):
                    out_dict[key]={'command':cls.numpy_type_to_str(val)}
                elif isinstance(val,np.ndarray):
                    out_dict[key]={'command':cls.numpy_to_str(val)}
                elif isinstance(val,slice):
                    out_dict[key]={'command': str(val)}
                elif isinstance(val,(list,tuple)):
                    out_dict[key]=cls.recursive_convert_list(val)
                elif isinstance(val, (dict,DictNamespace)):
                    out_dict[key]=cls.recoursive_convert_settings(val)
                else:
                    out_dict[key]=val
                    #raise ValueError('Cannot convert {} type for settings key {}'.format(type(val),key))
            except Exception as e:
                log.warning("Failed to save key {} and value type {} with error: \n {}".format(key,type(val),e))
                traceback.print_exc()
        return out_dict

    @classmethod
    def convert(cls,settings):
        return cls.recoursive_convert_settings(settings)



from xframe.library.mathLibrary import masked_mean,combine_means_2D
import numpy as np
from xframe.startup_routines import dependency_injection
dependency_injection()
from xframe.database.euxfel import default_DB
from xframe.presenters.matplolibPresenter import plot1D

db =default_DB()

data = db.load('/gpfs/exfel/theory_group/user/berberic/p3046/data/fig/run_32/2022-09-02_17:35:39/bg_data_proc.h5')


s = data['saxs']['saxs']
d = data['saxs']['polar_data']
m = data['saxs']['polar_mask'].astype(bool)
m[:]=d.astype(bool)
s2,counts = masked_mean(d,m,axis = 1)

fig1 = plot1D.get_fig(s)
fig2 = plot1D.get_fig(s2)


db.save('/gpfs/exfel/theory_group/user/berberic/p3046/data/fig/run_32/2022-09-02_17:35:39/s1.matplotlib',fig1)
db.save('/gpfs/exfel/theory_group/user/berberic/p3046/data/fig/run_32/2022-09-02_17:35:39/s2.matplotlib',fig2)
